import numpy as np
from mocap.dataset_loader import load_kuka_dataset
import pytransform3d.visualizer as pv
from pytransform3d.urdf import UrdfTransformManager
from movement_primitives.dmp import DualCartesianDMP


class SelectDemo:
    def __init__(self, fig, left, right):
        self.fig = fig
        self.left = left
        self.right = right
        self.show = True

    def __call__(self, vis, key, modifier):  # sometimes we don't receive the correct keys, why?
        self.show = not self.show
        if self.show and modifier:
            for g in self.left.geometries + self.right.geometries:
                vis.remove_geometry(g)
        elif not self.show and not modifier:
            for g in self.left.geometries + self.right.geometries:
                vis.add_geometry(g)
        fig.view_init(azim=0, elev=25)
        return True


fig = pv.figure(with_key_callbacks=True)
fig.plot_transform(s=0.1)
tm = UrdfTransformManager()
with open("kuka_lbr/urdf/kuka_lbr.urdf", "r") as f:
    tm.load_urdf(f.read(), mesh_path="kuka_lbr/urdf/")
fig.plot_graph(tm, "kuka_lbr", show_visuals=True)

n_weights_per_dim = 10

pattern = "data/kuka/20200129_peg_in_hole/csv_processed/01_peg_in_hole_both_arms/*.csv"
#pattern = "data/kuka/20191213_carry_heavy_load/csv_processed/01_heavy_load_no_tilt_0cm_dual_arm/*.csv"
#pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/panel_450mm_counterclockwise/*.csv"

dataset = load_kuka_dataset(pattern, verbose=1)

all_weights = []
all_starts = []
all_goals = []
all_execution_times = []

# first N trajectories are from training set
for idx in range(5):
    T, P = dataset[idx]

    execution_time = T[-1]
    dt = np.mean(np.diff(T))
    if dt < 0.005:  # HACK
        continue
    dmp = DualCartesianDMP(
        execution_time=execution_time, dt=dt,
        n_weights_per_dim=n_weights_per_dim, int_dt=0.01, p_gain=0.0)
    dmp.imitate(T, P)
    weights = dmp.get_weights()
    all_weights.append(weights)
    all_starts.append(P[0])
    all_goals.append(P[-1])
    all_execution_times.append(execution_time)

    if idx < 5:
        left = fig.plot_trajectory(P=P[:, :7], s=0.02)
        right = fig.plot_trajectory(P=P[:, 7:], s=0.02)
        key = ord(str((idx + 1) % 10))
        fig.visualizer.register_key_action_callback(key, SelectDemo(fig, left, right))

all_parameters = np.vstack([
    np.hstack((w, s, g, e)) for w, s, g, e in zip(
        all_weights, all_starts, all_goals, all_execution_times)])
mean_parameters = np.mean(all_parameters, axis=0)
cov_parameters = np.cov(all_parameters, rowvar=False)
random_state = np.random.RandomState(0)

# next 5 trajectories are sampled from Gaussian DMP
for idx in range(5):
    print("Reproduce #%d" % (idx + 1))

    params = random_state.multivariate_normal(mean_parameters, cov_parameters)
    n_weights = 2 * 6 * n_weights_per_dim
    weights = params[:n_weights]
    start = params[n_weights:n_weights + 14]
    goal = params[n_weights + 14:n_weights + 2 * 14]
    execution_time = params[-1]

    dmp = DualCartesianDMP(
        execution_time=execution_time, dt=0.01,
        n_weights_per_dim=n_weights_per_dim, int_dt=0.01)
    dmp.set_weights(weights)
    dmp.configure(start_y=start, goal_y=goal)
    T, P = dmp.open_loop(run_t=execution_time)

    left = fig.plot_trajectory(P=P[:, :7], s=0.02)
    right = fig.plot_trajectory(P=P[:, 7:], s=0.02)
    key = ord(str((idx + 6) % 10))
    fig.visualizer.register_key_action_callback(key, SelectDemo(fig, left, right))

fig.view_init(azim=0, elev=25)
fig.show()


"""
=======================
Minimum Jerk Trajectory
=======================

An example for a minimum jerk trajectory is displayed in the following plot.
"""
print(__doc__)

import matplotlib.pyplot as plt
from movement_primitives.data import generate_minimum_jerk

X, Xd, Xdd = generate_minimum_jerk([0], [1])
plt.figure()
plt.subplot(311)
plt.ylabel("$x$")
plt.plot(X[:, 0])
plt.subplot(312)
plt.ylabel("$\dot{x}$")
plt.plot(Xd[:, 0])
plt.subplot(313)
plt.xlabel("$t$")
plt.ylabel("$\ddot{x}$")
plt.plot(Xdd[:, 0])
plt.show()



"""Plotting Tools
==============

Tools for plotting multi-dimensional trajectories.
"""
import matplotlib.pyplot as plt


# extracted from seaborn to get rid of dependency:
# https://github.com/mwaskom/seaborn/blob/master/seaborn/palettes.py
PALETTE = ["#4C72B0", "#DD8452", "#55A868", "#C44E52", "#8172B3",
           "#937860", "#DA8BC3", "#8C8C8C", "#CCB974", "#64B5CD"]


def plot_trajectory_in_rows(
        trajectory, t=None, label=None, axes=None, subplot_shape=None,
        transpose=False, axis_titles=[], **kwargs):
    """Plot trajectories of N dimensions in N 2D subplots.

    Note that you have to manually activate the legend for one plot if you
    need it.

    Parameters
    ----------
    trajectory : array, shape (n_steps, n_dims)
        Trajectory

    t : array, shape (n_steps,), optional (default: step indices)
        Time of each step

    label : str, optional (default: None)
        Label that will appear in the legend for this trajectory

    axes : list
        Matplotlib axes on which the trajectory should be plotted in each
        dimension.

    subplot_shape : tuple
        Number of rows and number of columns.

    transpose : bool, optional (default: False)
        Fill first column first, then second column and so on. Typically,
        matplotlib fills rows before columns.

    axis_titles : list of str, optional (default: ['Dimension #0', ...])
        Title for each dimension of the plot.

    **kwargs : dict, optional
        Additional arguments for the plot function.

    Returns
    -------
    axes : list of matplotlib axis
        Matplotlib axes
    """
    n_steps, n_dims = trajectory.shape

    if subplot_shape is None:
        subplot_shape = (n_dims, 1)

    newaxes = axes is None
    if newaxes:
        axes = create_axes(n_dims, subplot_shape, transpose)

    if t is not None:
        xlabel = "Time [s]"
    else:
        t = range(n_steps)
        xlabel = "Step"

    for i in range(n_dims):
        axes[i].plot(t, trajectory[:, i], label=label, **kwargs)

    if newaxes:
        layout_axes(axes, n_dims, subplot_shape, xlabel, (t[0], t[-1]),
                    transpose, axis_titles)

    return axes


def plot_distribution_in_rows(
        mean, std_dev, t=None, label=None, axes=None, std_factors=(1, 2, 3),
        fill_between=True, subplot_shape=None, transpose=False, axis_titles=[],
        **kwargs):
    """Plot distribution of N dimensions in N 2D subplots.

    Note that you have to manually activate the legend for one plot if you
    need it.

    Parameters
    ----------
    mean : array, shape (n_steps, n_dims)
        Mean in each dimension

    std_dev : array, shape (n_steps, n_dims)
        Standard deviation in each dimension

    t : array, shape (n_steps,), optional (default: step indices)
        Time of each step

    label : str, optional (default: None)
        Label that will appear in the legend for this trajectory

    axes : list
        Matplotlib axes on which the trajectory should be plotted in each
        dimension.

    std_factors : tuple, optional (default: (1, 2, 3))
        Multiples of the standard deviation that will be indicates by
        area around the mean.

    fill_between : bool, optional (default: True)
        Fill area around mean. Multiples of standard deviation will be
        indicated by dashed lines otherwise.

    subplot_shape : tuple
        Number of rows and number of columns.

    transpose : bool, optional (default: False)
        Fill first column first, then second column and so on. Typically,
        matplotlib fills rows before columns.

    axis_titles : list of str, optional (default: ['Dimension #0', ...])
        Title for each dimension of the plot.

    **kwargs : dict, optional
        Additional arguments for the plot function.

    Returns
    -------
    axes : list of matplotlib axis
        Matplotlib axes
    """
    n_steps, n_dims = mean.shape

    if subplot_shape is None:
        subplot_shape = (n_dims, 1)

    newaxes = axes is None
    if newaxes:
        axes = create_axes(n_dims, subplot_shape, transpose)

    if t is None:
        t = range(n_steps)
        xlabel = "Step"
    else:
        xlabel = "Time [s]"

    color = kwargs.get("color", None)
    alpha = kwargs.get("alpha", 0.1)

    for i in range(n_dims):
        axes[i].plot(t, mean[:, i], **kwargs)
        for f_idx, f in enumerate(std_factors):
            if fill_between:
                axes[i].fill_between(
                    t, mean[:, i] - f * std_dev[:, i],
                    mean[:, i] + f * std_dev[:, i], color=color, alpha=alpha,
                    label=label if f_idx == 0 else None)
            else:
                axes[i].plot(t, mean[:, i] - f * std_dev[:, i], color, ls="--",
                             label=label if f_idx == 0 else None)
                axes[i].plot(t, mean[:, i] + f * std_dev[:, i], color, ls="--")

    if newaxes:
        layout_axes(axes, n_dims, subplot_shape, xlabel, (t[0], t[-1]),
                    transpose, axis_titles)

    return axes


def create_axes(n_dims, subplot_shape, transpose):
    """Create multiple axes to plot trajectories.

    Parameters
    ----------
    n_dims : int
        Number of dimensions to plot.

    subplot_shape : tuple
        Desired rows and columns of the figure.

    transpose : bool
        Switch rows and columns?

    Returns
    -------
    axes : list of matplotlib axes
        Matplotlib axes.
    """
    h, w = subplot_shape
    if transpose:
        dim_order = [(i % h) * w + i // h for i in range(n_dims)]
    else:
        dim_order = range(n_dims)
    return [plt.subplot(h, w, 1 + i) for i in dim_order]


def layout_axes(axes, n_dims, subplot_shape, xlabel, xlim, transpose,
                axis_titles):
    for i in range(n_dims):
        if i < len(axis_titles):
            axis_title = axis_titles[i]
        else:
            axis_title = "Dimension #%d" % i
        axes[i].set_title(axis_title, loc="center", y=0,
                          backgroundcolor="#ffffff80")
        if not transpose and subplot_shape[0] * subplot_shape[1] - i in range(
                1, subplot_shape[1] + 1):
            axes[i].set_xlabel(xlabel)
        elif transpose and i % subplot_shape[0] == subplot_shape[0] - 1:
            axes[i].set_xlabel(xlabel)
        else:
            axes[i].set_xticks(())
        axes[i].set_xlim(xlim)
    plt.tight_layout(h_pad=0)



import numpy as np
from ._base import DMPBase, WeightParametersMixin
from ..utils import ensure_1d_array
from ._forcing_term import ForcingTerm, phase
from ._canonical_system import canonical_system_alpha
from ._dmp import dmp_imitate, dmp_open_loop


class DMPWithFinalVelocity(WeightParametersMixin, DMPBase):
    r"""Dynamical movement primitive (DMP) with final velocity.

    Equation of transformation system (according to [1]_, Eq. 6):

    .. math::

        \ddot{y} = (\alpha_y (\beta_y (g - y) + \tau\dot{g} - \tau \dot{y}) + f(z))/\tau^2

    Parameters
    ----------
    n_dims : int
        State space dimensions.

    execution_time : float, optional (default: 1)
        Execution time of the DMP: :math:`\tau`.

    dt : float, optional (default: 0.01)
        Time difference between DMP steps: :math:`\Delta t`.

    n_weights_per_dim : int, optional (default: 10)
        Number of weights of the function approximator per dimension.

    int_dt : float, optional (default: 0.001)
        Time difference for Euler integration.

    p_gain : float, optional (default: 0)
        Gain for proportional controller of DMP tracking error.
        The domain is [0, execution_time**2/dt].

    alpha_y : float or array-like, shape (n_dims,), optional (default: 25.0)
        Parameter of the transformation system.

    beta_y : float or array-like, shape (n_dims,), optional (default: 6.25)
        Parameter of the transformation system.

    Attributes
    ----------
    execution_time_ : float
        Execution time of the DMP.

    dt_ : float
        Time difference between DMP steps. This value can be changed to adapt
        the frequency.

    References
    ----------
    .. [1] Muelling, K., Kober, J., Kroemer, O., Peters, J. (2013). Learning to
       Select and Generalize Striking Movements in Robot Table Tennis.
       International Journal of Robotics Research 32 (3), 263-279.
       https://www.ias.informatik.tu-darmstadt.de/uploads/Publications/Muelling_IJRR_2013.pdf
    """
    def __init__(self, n_dims, execution_time=1.0, dt=0.01,
                 n_weights_per_dim=10, int_dt=0.001, p_gain=0.0,
                 alpha_y=25.0, beta_y=6.25):
        super(DMPWithFinalVelocity, self).__init__(n_dims, n_dims)
        self._execution_time = execution_time
        self.dt_ = dt
        self.n_weights_per_dim = n_weights_per_dim
        self.int_dt = int_dt
        self.p_gain = p_gain
        self.smooth_scaling = False  # it is build into this DMP type already

        self._init_forcing_term()

        self.alpha_y = ensure_1d_array(alpha_y, n_dims, "alpha_y")
        self.beta_y = ensure_1d_array(beta_y, n_dims, "beta_y")

    def _init_forcing_term(self):
        alpha_z = canonical_system_alpha(0.01, self.execution_time_, 0.0)
        self.forcing_term = ForcingTerm(
            self.n_dims, self.n_weights_per_dim, self.execution_time_,
            0.0, 0.8, alpha_z)

    def get_execution_time_(self):
        return self._execution_time

    def set_execution_time_(self, execution_time):
        self._execution_time = execution_time
        weights = self.forcing_term.weights_
        self._init_forcing_term()
        self.forcing_term.weights_ = weights

    execution_time_ = property(get_execution_time_, set_execution_time_)

    def step(self, last_y, last_yd, coupling_term=None):
        """DMP step.

        Parameters
        ----------
        last_y : array, shape (n_dims,)
            Last state.

        last_yd : array, shape (n_dims,)
            Last time derivative of state (e.g., velocity).

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        Returns
        -------
        y : array, shape (n_dims,)
            Next state.

        yd : array, shape (n_dims,)
            Next time derivative of state (e.g., velocity).
        """
        self.last_t = self.t
        self.t += self.dt_

        if not self.initialized:
            self.current_y = np.copy(self.start_y)
            self.current_yd = np.copy(self.start_yd)
            self.initialized = True

        # https://github.com/studywolf/pydmps/blob/master/pydmps/cs.py
        tracking_error = self.current_y - last_y

        dmp_step_euler_with_constraints(
            self.last_t, self.t,
            self.current_y, self.current_yd,
            self.goal_y, self.goal_yd, self.goal_ydd,
            self.start_y, self.start_yd, self.start_ydd,
            self.execution_time_, 0.0,
            self.alpha_y, self.beta_y,
            self.forcing_term,
            coupling_term=coupling_term,
            int_dt=self.int_dt,
            p_gain=self.p_gain,
            tracking_error=tracking_error,
            smooth_scaling=self.smooth_scaling)
        return np.copy(self.current_y), np.copy(self.current_yd)

    def open_loop(self, run_t=None, coupling_term=None):
        """Run DMP open loop.

        Parameters
        ----------
        run_t : float, optional (default: execution_time)
            Run time of DMP. Can be shorter or longer than execution_time.

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        Returns
        -------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, n_dims)
            State at each step.
        """
        return dmp_open_loop(
            self.execution_time_, 0.0, self.dt_,
            self.start_y, self.goal_y,
            self.alpha_y, self.beta_y,
            self.forcing_term,
            coupling_term,
            run_t, self.int_dt,
            dmp_step_euler_with_constraints,
            start_yd=self.start_yd, start_ydd=self.start_ydd,
            goal_yd=self.goal_yd, goal_ydd=self.goal_ydd,
            smooth_scaling=self.smooth_scaling)

    def imitate(self, T, Y, regularization_coefficient=0.0):
        r"""Imitate demonstration.

        Target forces of the forcing term are computed according to

        .. math::

            f_{target} =
            \tau^2 \ddot{y}_{demo}
            - \alpha_y(
                \beta_y (g-y_{demo})
                + \tau (\dot{g} - \dot{y}_{demo})
            )
            - \tau^2 \ddot{g},

        where :math:`g, \dot{g}, \ddot{g}` are constraints that will be
        recomputed in each step.

        Parameters
        ----------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, n_dims)
            State at each step.

        regularization_coefficient : float, optional (default: 0)
            Regularization coefficient for regression.
        """
        self.forcing_term.weights_[:, :], start_y, start_yd, \
            start_ydd, goal_y, goal_yd, goal_ydd = dmp_imitate(
            T, Y,
            n_weights_per_dim=self.n_weights_per_dim,
            regularization_coefficient=regularization_coefficient,
            alpha_y=self.alpha_y, beta_y=self.beta_y,
            overlap=self.forcing_term.overlap,
            alpha_z=self.forcing_term.alpha_z, allow_final_velocity=True,
            determine_forces=determine_forces)
        self.configure(
            start_y=start_y, start_yd=start_yd, start_ydd=start_ydd,
            goal_y=goal_y, goal_yd=goal_yd, goal_ydd=goal_ydd)


def solve_constraints(t0, t1, y0, y0d, y0dd, y1, y1d, y1dd):
    t02 = t0 * t0
    t03 = t02 * t0
    t04 = t03 * t0
    t05 = t04 * t0
    t12 = t1 * t1
    t13 = t12 * t1
    t14 = t13 * t1
    t15 = t14 * t1

    M = np.array([[1, t0, t02, t03, t04, t05],
                  [0, 1, 2 * t0, 3 * t02, 4 * t03, 5 * t04],
                  [0, 0, 2, 6 * t0, 12 * t02, 20 * t03],
                  [1, t1, t12, t13, t14, t15],
                  [0, 1, 2 * t1, 3 * t12, 4 * t13, 5 * t14],
                  [0, 0, 2, 6 * t1, 12 * t12, 20 * t13]])
    Y = np.vstack((y0, y0d, y0dd, y1, y1d, y1dd))

    # Solve M*b = y for b in each DOF at once
    B = np.linalg.solve(M, Y)
    return B


def apply_constraints(t, goal_y, goal_t, coefficients):
    if t > goal_t + np.finfo(float).eps:
        # For t > goal_t the polynomial should always 'pull' to the goal
        # position, but velocity and acceleration should be zero.
        # This is done to avoid diverging from the goal if the DMP is executed
        # longer than expected.
        return goal_y, np.zeros_like(goal_y), np.zeros_like(goal_y)
    else:
        t2 = t * t
        t3 = t2 * t
        t4 = t3 * t
        t5 = t4 * t
        pos = np.array([1, t, t2, t3, t4, t5])
        vel = np.array([0, 1, 2 * t, 3 * t2, 4 * t3, 5 * t4])
        acc = np.array([0, 0, 2, 6 * t, 12 * t2, 20 * t3])

        g = np.dot(pos, coefficients)
        gd = np.dot(vel, coefficients)
        gdd = np.dot(acc, coefficients)
        return g, gd, gdd


def determine_forces(T, Y, alpha_y, beta_y, alpha_z, allow_final_velocity,
                     smooth_scaling=False):
    """Determine forces that the forcing term should generate.

    Parameters
    ----------
    T : array, shape (n_steps,)
        Time of each step.

    Y : array, shape (n_steps, n_dims)
        Position at each step.

    alpha_y : array, shape (n_dims,)
        Parameter of the transformation system.

    beta_y : array, shape (n_dims,)
        Parameter of the transformation system.

    alpha_z : float
        Parameter of the canonical system.

    allow_final_velocity : bool
        Whether a final velocity is allowed. This should always be True for
        this function.

    smooth_scaling : bool, optional (default: False)
        Not used. This DMP type adapts smoothly to goal changes as constraint
        coefficients will be recomputed in each step.

    Returns
    -------
    F : array, shape (n_steps, n_dims)
        Forces.

    start_y : array, shape (n_dims,)
        Start position.

    start_yd : array, shape (n_dims,)
        Start velocity.

    start_ydd : array, shape (n_dims,)
        Start acceleration.

    goal_y : array, shape (n_dims,)
        Final position.

    goal_yd : array, shape (n_dims,)
        Final velocity.

    goal_ydd : array, shape (n_dims,)
        Final acceleration.
    """
    assert allow_final_velocity

    n_dims = Y.shape[1]
    DT = np.diff(T)

    Yd = np.empty_like(Y)
    Yd[0] = 0.0
    for d in range(n_dims):
        Yd[1:, d] = np.diff(Y[:, d]) / DT

    Ydd = np.empty_like(Y)
    Ydd[0] = 0.0
    for d in range(n_dims):
        Ydd[1:, d] = np.diff(Yd[:, d]) / DT

    coefficients = solve_constraints(
        T[0], T[-1], Y[0], Yd[0], Ydd[0], Y[-1], Yd[-1], Ydd[-1])

    execution_time = T[-1] - T[0]
    F = np.empty((len(T), n_dims))
    for t in range(len(T)):
        g, gd, gdd = apply_constraints(T[t], Y[-1], T[-1], coefficients)
        F[t, :] = execution_time ** 2 * Ydd[t] - alpha_y * (
            beta_y * (g - Y[t])
            + execution_time * (gd - Yd[t])
        ) - execution_time ** 2 * gdd
    return F, Y[0], Yd[0], Ydd[0], Y[-1], Yd[-1], Ydd[-1]


def dmp_step_euler_with_constraints(
        last_t, t, current_y, current_yd, goal_y, goal_yd, goal_ydd,
        start_y, start_yd, start_ydd, goal_t, start_t, alpha_y, beta_y,
        forcing_term, coupling_term=None, coupling_term_precomputed=None,
        int_dt=0.001, p_gain=0.0, tracking_error=0.0, smooth_scaling=False):
    """Integrate regular DMP for one step with Euler integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (n_dims,)
        Current position. Will be modified.

    current_yd : array, shape (n_dims,)
        Current velocity. Will be modified.

    goal_y : array, shape (n_dims,)
        Goal position.

    goal_yd : array, shape (n_dims,)
        Goal velocity.

    goal_ydd : array, shape (n_dims,)
        Goal acceleration.

    start_y : array, shape (n_dims,)
        Start position.

    start_yd : array, shape (n_dims,)
        Start velocity.

    start_ydd : array, shape (n_dims,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : array, shape (n_dims,)
        Constant in transformation system.

    beta_y : array, shape (n_dims,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    coupling_term_precomputed : tuple
        A precomputed coupling term, i.e., additional velocity and
        acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    p_gain : float, optional (default: 0)
        Proportional gain for tracking error.

    tracking_error : float, optional (default: 0)
        Tracking error from last step.

    smooth_scaling : bool, optional (default: False)
        Not used. This DMP type adapts smoothly to goal changes as constraint
        coefficients will be recomputed in each step.
    """
    if start_t >= goal_t:
        raise ValueError("Goal must be chronologically after start!")

    if t <= start_t:
        return np.copy(start_y), np.copy(start_yd), np.copy(start_ydd)

    execution_time = goal_t - start_t

    coefficients = solve_constraints(
        start_t, goal_t, start_y, start_yd, start_ydd,
        goal_y, goal_yd, goal_ydd)

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is not None:
            cd, cdd = coupling_term.coupling(current_y, current_yd)
        else:
            cd, cdd = np.zeros_like(current_y), np.zeros_like(current_y)
        if coupling_term_precomputed is not None:
            cd += coupling_term_precomputed[0]
            cdd += coupling_term_precomputed[1]

        z = forcing_term.phase(current_t)
        f = forcing_term.forcing_term(z).squeeze()

        g, gd, gdd = apply_constraints(current_t, goal_y, goal_t, coefficients)

        coupling_sum = cdd + p_gain * tracking_error / dt
        ydd = (
            alpha_y * (
                beta_y * (g - current_y)
                + execution_time * (gd - current_yd)
            )
            + f
            + coupling_sum
        ) / execution_time ** 2 + gdd
        current_yd += dt * ydd + cd / execution_time
        current_y += dt * current_yd



"""Input / Output
==============

Input and output from and to files of movement primitives.
"""
import inspect
import json
import pickle
import yaml
import numpy as np


basic_types = (int, float, bool, str, type(None))
basic_types_and_sequences = (int, float, bool, str, list, tuple, type(None))


def write_pickle(filename, obj):
    """Write object to pickle format.

    Parameters
    ----------
    filename : str
        Output file.

    obj : object
        Any object.
    """
    with open(filename, "wb") as f:
        pickle.dump(obj, f)


def read_pickle(filename):
    """Read object from pickle format.

    Parameters
    ----------
    filename : str
        Input file.

    Returns
    -------
    obj : object
        Python object.
    """
    with open(filename, "rb") as f:
        return pickle.load(f)


def write_yaml(filename, obj):
    """Write object to YAML format.

    Parameters
    ----------
    filename : str
        Output file.

    obj : object
        Any custom object that is a hierarchical composition of basic data
        types and numpy arrays.
    """
    export = _recursive_to_dict(obj, True)
    with open(filename, "w") as f:
        yaml.dump(export, f)


def read_yaml(filename):
    """Read object from YAML format.

    Parameters
    ----------
    filename : str
        Input file.

    Returns
    -------
    obj : object
        Python object.
    """
    with open(filename, "r") as f:
        export = yaml.safe_load(f)
    return _dict_to_object(export)


def write_json(filename, obj):
    """Write object to JSON format.

    Parameters
    ----------
    filename : str
        Output file.

    obj : object
        Any custom object that is a hierarchical composition of basic data
        types and numpy arrays.
    """
    export = _recursive_to_dict(obj)
    with open(filename, "w") as f:
        json.dump(export, f)


def read_json(filename):
    """Read object from JSON format.

    Parameters
    ----------
    filename : str
        Input file.

    Returns
    -------
    obj : object
        Python object.
    """
    with open(filename, "r") as f:
        export = json.load(f)
    return _dict_to_object(export)


def _recursive_to_dict(obj, convert_tuple=False):
    result = {"module": obj.__module__, "class": obj.__class__.__name__}
    for k, v in obj.__dict__.items():
        if convert_tuple and isinstance(v, tuple):
            result[k] = list(v)
        elif isinstance(v, basic_types_and_sequences):
            result[k] = v
        elif isinstance(v, np.ndarray):
            result[k] = v.tolist()
        else:
            result[k] = _recursive_to_dict(v)
    return result


def _recursive_from_dict(obj, export):
    for k, v in export.items():
        if isinstance(v, basic_types):
            setattr(obj, k, v)
        elif isinstance(v, (tuple, list)):
            if isinstance(obj.__dict__[k], np.ndarray):
                obj.__dict__[k] = np.array(v)
            else:
                setattr(obj, k, v)
        else:
            _recursive_from_dict(getattr(obj, k), v)


def _dict_to_object(export):
    module_name = export.pop("module")
    module = __import__(module_name, {}, {}, fromlist=["dummy"], level=0)
    class_dict = dict(inspect.getmembers(module))
    class_name = export.pop("class")
    if class_name not in class_dict:
        raise ImportError(f"cannot import name '{class_name}' from '{module}'")
    clazz = class_dict[class_name]

    argspec = inspect.getfullargspec(clazz)
    ctor_kwargs = {}
    for arg in argspec.args:
        if arg in export:
            ctor_kwargs[arg] = export.pop(arg)
    obj = clazz(**ctor_kwargs)

    _recursive_from_dict(obj, export)
    return obj



import numpy as np
import matplotlib.pyplot as plt
from mocap.dataset_loader import load_kuka_dataset
from movement_primitives.plot import plot_trajectory_in_rows
from mocap.cleaning import median_filter
from pytransform3d.batch_rotations import smooth_quaternion_trajectory
from pytransform3d import trajectories as ptr


n_weights_per_dim = 10

#pattern = "data/kuka/20200129_peg_in_hole/csv_processed/01_peg_in_hole_both_arms/*.csv"
#pattern = "data/kuka/20191213_carry_heavy_load/csv_processed/01_heavy_load_no_tilt_0cm_dual_arm/*.csv"
#pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/panel_450mm_counterclockwise/*.csv"
#pattern = "data/kuka/20191213_carry_heavy_load/csv_processed/*/*.csv"
pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/*counterclockwise/*.csv"

dataset = load_kuka_dataset(pattern, verbose=1)

plt.figure()
axes = None
for T, P in dataset:
    P_left = P[:, :7]
    P_right = P[:, 7:]

    P_left[:, 3:] = smooth_quaternion_trajectory(P_left[:, 3:])
    P_right[:, 3:] = smooth_quaternion_trajectory(P_right[:, 3:])
    P_left[:, :] = median_filter(P_left, window_size=5)
    P_right[:, :] = median_filter(P_right, window_size=5)

    left2base = ptr.transforms_from_pqs(P_left)
    right2base = ptr.transforms_from_pqs(P_right)

    left2right = np.empty_like(left2base)
    for t in range(len(left2right)):
        left2right[t] = left2base[t].dot(np.linalg.inv(right2base[t]))
    P_diff = ptr.pqs_from_transforms(left2right)

    P_diff[:, 3:] = smooth_quaternion_trajectory(P_diff[:, 3:])
    P_diff[:, :] = median_filter(P_diff, window_size=5)

    P = np.hstack((P_left, P_right, P_diff))

    axes = plot_trajectory_in_rows(P, T, axes=axes, subplot_shape=(7, 3), transpose=True)
plt.tight_layout()
plt.show()



"""Data
====

Tools for loading datasets.
"""
from ._lasa import load_lasa
from ._minimum_jerk import generate_minimum_jerk
from ._toy_1d import generate_1d_trajectory_distribution


__all__ = [
    "load_lasa",
    "generate_minimum_jerk",
    "generate_1d_trajectory_distribution"]



# git clone git@git.hb.dfki.de:models-robots/rh5v2_models/pybullet-urdf.git --branch develop --recursive
import numpy as np
import pybullet
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
import pytransform3d.trajectories as ptr
from pytransform3d.batch_rotations import smooth_quaternion_trajectory
from movement_primitives.dmp import DualCartesianDMP
from movement_primitives.kinematics import Kinematics
from movement_primitives.testing.simulation import RH5Simulation, draw_transform, _pybullet_pose, get_absolute_path
from mocap.cleaning import smooth_exponential_coordinates
from movement_primitives.io import write_json, write_yaml, write_pickle


def panel_pose(tcp_left, tcp_right):
    tcp_left_pos = tcp_left[:3, 3]
    tcp_right_pos = tcp_right[:3, 3]
    tcp_middle = 0.5 * (tcp_left_pos + tcp_right_pos)
    tcp_middle[0] += 0.025
    x_axis = pr.norm_vector(tcp_right_pos - tcp_left_pos)
    y_axis = -pr.norm_vector(0.5 * (tcp_left[:3, 2] + tcp_right[:3, 2]))
    R_panel = pr.matrix_from_two_vectors(x_axis, y_axis)
    return pt.transform_from(R_panel, tcp_middle)


panel_rotation_angle = np.deg2rad(30)
n_steps = 1001
dt = 0.001

#q0 = np.array([-1.57, 1.25, 0, -1.75, 0, 0, 0.8, -1.57, 1.25, 0, -1.75, 0, 0, 0.8])
#q0 = np.array([-1.57, 0.76, 0, -1.3, 0, 0, -0.55, -1.57, 0.76, 0, -1.3, 0, 0, -0.55])
# q0 = np.array([-1.57, 0.9, 0, -1.05, 0, 0, 0, -1.57, 0.9, 0, -1.05, -0.25, 0, 0])
# new position for closed gripper
q0 = np.array([-1.59, 0.81, 0, -1.11, -0.25, 0.055, 0.56, -1.59, 0.81, 0, -1.11, -0.25, -0.055, -0.56])

rh5 = RH5Simulation(dt=dt, gui=True, real_time=False,
                    urdf_path=get_absolute_path("pybullet-urdf/urdf/RH5v2.urdf", "models/robots/rh5v2_models"),
                    left_arm_path=get_absolute_path("pybullet-urdf/submodels/left_arm.urdf", "models/robots/rh5v2_models"),
                    right_arm_path=get_absolute_path("pybullet-urdf/submodels/right_arm.urdf", "models/robots/rh5v2_models"))
rh5.set_desired_joint_state(q0, position_control=True)
rh5.sim_loop(int(1.0 / dt))
ee_state = rh5.get_ee_state()
left2base_start = pt.transform_from_pq(ee_state[:7])
right2base_start = pt.transform_from_pq(ee_state[7:])
draw_transform(left2base_start, s=0.1, client_id=rh5.client_id)
draw_transform(right2base_start, s=0.1, client_id=rh5.client_id)

panel2base_start = panel_pose(left2base_start, right2base_start)
draw_transform(panel2base_start, s=0.1, client_id=rh5.client_id)
panel2base_start_pq = pt.pq_from_transform(panel2base_start)
p, q = _pybullet_pose(panel2base_start_pq)

pybullet.loadURDF(get_absolute_path("solar_panels/solar_panel_02/urdf/pb_solar_panel_02.urdf", "models/objects"), p, q)
#time.sleep(10)

left2panel_start = pt.concat(left2base_start, pt.invert_transform(panel2base_start))
right2panel_start = pt.concat(right2base_start, pt.invert_transform(panel2base_start))

rotation_axis = -pr.unity
start2end = pt.rotate_transform(np.eye(4), pr.matrix_from_compact_axis_angle(rotation_axis * panel_rotation_angle))
left2panel_end = pt.concat(left2panel_start, start2end)
right2panel_end = pt.concat(right2panel_start, start2end)

left2base_end = pt.concat(left2panel_end, panel2base_start)
right2base_end = pt.concat(right2panel_end, panel2base_start)

start_left = pt.exponential_coordinates_from_transform(left2base_start)
end_left = pt.exponential_coordinates_from_transform(left2base_end)
start_right = pt.exponential_coordinates_from_transform(right2base_start)
end_right = pt.exponential_coordinates_from_transform(right2base_end)
start_left, end_left = smooth_exponential_coordinates(np.array([start_left, end_left]))
start_right, end_right = smooth_exponential_coordinates(np.array([start_right, end_right]))

t = np.linspace(0, 1, n_steps)
left_trajectory = start_left[np.newaxis] + t[:, np.newaxis] * (end_left[np.newaxis] - start_left[np.newaxis])
right_trajectory = start_right[np.newaxis] + t[:, np.newaxis] * (end_right[np.newaxis] - start_right[np.newaxis])
left_trajectory = ptr.transforms_from_exponential_coordinates(left_trajectory)
right_trajectory = ptr.transforms_from_exponential_coordinates(right_trajectory)

########################################################################################################################
# Data export
with open(get_absolute_path("pybullet-urdf/urdf/RH5v2.urdf", "models/robots/rh5v2_models"), "r") as f:
    kin = Kinematics(f.read(), mesh_path=get_absolute_path("pybullet-urdf/urdf/", "models/robots/rh5v2_models"))
#kin.tm.write_png("graph.png", "twopi")

lwp2ltcp = kin.tm.get_transform("ALWristFT_Link", "LTCP_Link")
rwp2rtcp = kin.tm.get_transform("ARWristFT_Link", "RTCP_Link")

lwp_trajectory = np.array([pt.concat(lwp2ltcp, ltcp2base) for ltcp2base in left_trajectory])
rwp_trajectory = np.array([pt.concat(rwp2rtcp, rtcp2base) for rtcp2base in right_trajectory])

left_trajectory_pq = ptr.pqs_from_transforms(lwp_trajectory)
right_trajectory_pq = ptr.pqs_from_transforms(rwp_trajectory)

dmp = DualCartesianDMP(execution_time=n_steps * dt, dt=dt, n_weights_per_dim=10)
Y_pq = np.empty((n_steps, 14))
Y_pq[:, :7] = left_trajectory_pq
Y_pq[:, 7:] = right_trajectory_pq
Y_pq[:, 3:7] = smooth_quaternion_trajectory(Y_pq[:, 3:7])
Y_pq[:, 10:14] = smooth_quaternion_trajectory(Y_pq[:, 10:14])
T = np.linspace(0, 1, n_steps)
dmp.imitate(T, Y_pq)

write_yaml("rh5v2_dual_arm_dmp.yaml", dmp)
write_json("rh5v2_dual_arm_dmp.json", dmp)
write_pickle("rh5v2_dual_arm_dmp.pickle", dmp)
########################################################################################################################

#draw_trajectory(left_trajectory, rh5.client_id, s=0.05, n_key_frames=5)
#draw_trajectory(right_trajectory, rh5.client_id, s=0.05, n_key_frames=5)

print("Imitation...")
T = np.linspace(0, (n_steps - 1) * dt, n_steps)
P = np.hstack((ptr.pqs_from_transforms(left_trajectory),
               ptr.pqs_from_transforms(right_trajectory)))
dmp = DualCartesianDMP(execution_time=T[-1], dt=dt, n_weights_per_dim=10)
dmp.imitate(T, P)

desired_positions, positions, desired_velocities, velocities = \
    rh5.step_through_cartesian(dmp, P[0], np.zeros(12), T[-1])

rh5.stop()
rh5.sim_loop()



import os
import zipfile
import io
import scipy.io
import numpy as np
try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen


LASA_URL = ("http://bitbucket.org/khansari/lasahandwritingdataset/get/"
            "38304f7c0ac4.zip")


def load_lasa(shape_idx):
    """Load demonstrations from LASA dataset.

    The LASA dataset contains 2D handwriting motions recorded from a
    Tablet-PC. It can be found `here
    <https://bitbucket.org/khansari/lasahandwritingdataset>`_.
    Take a look at the `detailed explanation
    <https://cs.stanford.edu/people/khansari/download.html>`_
    for more information. It was initially used to evaluate stable estimators
    of dynamical systems [1]_.

    Parameters
    ----------
    shape_idx : int
        Choose demonstrated shape, must be within range(30).

    Returns
    -------
    T : array, shape (n_demos, n_steps)
        Times

    X : array, shape (n_demos, n_steps, n_dims)
        Positions

    Xd : array, shape (n_demos, n_steps, n_dims)
        Velocities

    Xdd : array, shape (n_demos, n_steps, n_dims)
        Accelerations

    dt : float
        Time between steps

    shape_name : string
        Name of the Matlab file from which we load the demonstrations
        (without suffix).

    References
    ----------
    .. [1] Khansari-Zadeh, S. M., Billard, A. (2011).
       Learning Stable Nonlinear Dynamical Systems With Gaussian Mixture Models.
       IEEE Transactions on Robotics, 27 (5), 943-957. DOI:
       10.1109/TRO.2011.2159412
    """
    dataset_path = get_common_dataset_path()
    if not os.path.isdir(dataset_path + "lasa_data"):  # pragma: no cover
        url = urlopen(LASA_URL)
        z = zipfile.ZipFile(io.BytesIO(url.read()))
        z.extractall(dataset_path)
        os.rename(dataset_path + z.namelist()[0],
                  dataset_path + "lasa_data" + os.sep)

    dataset_path += "lasa_data" + os.sep + "DataSet" + os.sep
    demos, shape_name = _load_from_matlab_file(dataset_path, shape_idx)
    X, Xd, Xdd, dt = _convert_demonstrations(demos)
    t = np.linspace(0, X.shape[1], X.shape[1])
    T = np.tile(t, (X.shape[0], 1))
    return T, X, Xd, Xdd, dt, shape_name


def _load_from_matlab_file(dataset_path, shape_idx):
    """Load demonstrations from Matlab files.

    Parameters
    ----------
    dataset_path : str
        Path where external datasets are stored.

    shape_idx : int
        Index of the shape in the LASA dataset.

    Returns
    -------
    demos : array
        Demonstrations.

    shape_name : string
        Name of the Matlab file from which we load the demonstrations
        (without suffix).
    """
    file_name = sorted(os.listdir(dataset_path))[shape_idx]
    return (scipy.io.loadmat(dataset_path + file_name)["demos"][0],
            file_name[:-4])


def _convert_demonstrations(demos):
    """Convert Matlab struct to numpy arrays.

    Parameters
    ----------
    demos : array
        Demonstrations.

    Returns
    -------
    X : array, shape (n_demos, n_steps, n_dims)
        Positions

    Xd : array, shape (n_demos, n_steps, n_dims)
        Velocities

    Xdd : array, shape (n_demos, n_steps, n_dims)
        Accelerations

    dt : float
        Time between steps
    """
    tmp = []
    for demo_idx in range(demos.shape[0]):
        # The Matlab format is strange...
        demo = demos[demo_idx][0, 0]
        # Positions, velocities and accelerations
        tmp.append((demo[0], demo[2], demo[3]))

    X = np.transpose([P for P, _, _ in tmp], [0, 2, 1])
    Xd = np.transpose([V for _, V, _ in tmp], [0, 2, 1])
    Xdd = np.transpose([A for _, _, A in tmp], [0, 2, 1])

    dt = float(demos[0][0, 0][4])

    return X, Xd, Xdd, dt


def get_common_dataset_path():
    """Returns the path where all external datasets are stored.

    Returns
    -------
    dataset_path : str
        Path where external datasets are stored.
    """
    dataset_path = os.path.expanduser("~")
    dataset_path += os.sep + ".movement_primitive_data" + os.sep
    return dataset_path



import numpy as np
import open3d as o3d
from matplotlib import cbook
import pytransform3d.visualizer as pv
import pytransform3d.transformations as pt
import pytransform3d.trajectories as ptr
import pytransform3d.uncertainty as pu
from pytransform3d.urdf import UrdfTransformManager
from mocap.cleaning import smooth_exponential_coordinates, median_filter
from mocap.dataset_loader import load_kuka_dataset, transpose_dataset

from movement_primitives.promp import ProMP
from gmr import GMM


class Surface(pv.Artist):
    """Surface.

    Parameters
    ----------
    x : array, shape (n_steps, n_steps)
        Coordinates on x-axis of grid on surface.
    y : array, shape (n_steps, n_steps)
        Coordinates on y-axis of grid on surface.
    z : array, shape (n_steps, n_steps)
        Coordinates on z-axis of grid on surface.
    c : array-like, shape (3,), optional (default: None)
        Color
    """
    def __init__(self, x, y, z, c=None):
        self.c = c
        self.mesh = o3d.geometry.TriangleMesh()
        self.set_data(x, y, z)

    def set_data(self, x, y, z):
        """Update data.

        Parameters
        ----------
        x : array, shape (n_steps, n_steps)
            Coordinates on x-axis of grid on surface.
        y : array, shape (n_steps, n_steps)
            Coordinates on y-axis of grid on surface.
        z : array, shape (n_steps, n_steps)
            Coordinates on z-axis of grid on surface.
        """
        polys = np.stack([cbook._array_patch_perimeters(a, 1, 1)
                          for a in (x, y, z)], axis=-1)
        vertices = polys.reshape(-1, 3)
        triangles = (
            [[4 * i + 0, 4 * i + 1, 4 * i + 2] for i in range(len(polys))] +
            [[4 * i + 2, 4 * i + 3, 4 * i + 0] for i in range(len(polys))] +
            [[4 * i + 0, 4 * i + 3, 4 * i + 2] for i in range(len(polys))] +
            [[4 * i + 2, 4 * i + 1, 4 * i + 0] for i in range(len(polys))]
        )
        self.mesh.vertices = o3d.utility.Vector3dVector(vertices)
        self.mesh.triangles = o3d.utility.Vector3iVector(triangles)
        if self.c is not None:
            self.mesh.paint_uniform_color(self.c)
        self.mesh.compute_vertex_normals()

    @property
    def geometries(self):
        """Expose geometries.
        Returns
        -------
        geometries : list
            List of geometries that can be added to the visualizer.
        """
        return [self.mesh]


def generate_training_data(
        pattern, n_weights_per_dim, context_names, verbose=0):
    Ts, Ps, contexts = transpose_dataset(
        load_kuka_dataset(pattern, context_names, verbose=verbose))

    Es = []
    for P in Ps:
        E = np.empty((len(P), 2 * 6))
        E[:, :6] = ptr.exponential_coordinates_from_transforms(ptr.transforms_from_pqs(P[:, :7]))
        E[:, 6:] = ptr.exponential_coordinates_from_transforms(ptr.transforms_from_pqs(P[:, 7:]))
        E[:, :6] = smooth_exponential_coordinates(E[:, :6])
        E[:, 6:] = smooth_exponential_coordinates(E[:, 6:])
        E[:, :] = median_filter(E, 5)
        Es.append(E)

        #import matplotlib.pyplot as plt
        #from movement_primitives.plot import plot_trajectory_in_rows
        #plot_trajectory_in_rows(E, subplot_shape=(6, 2))
        #plt.show()

    n_demos = len(Ts)
    n_dims = Es[0].shape[1]

    promp = ProMP(n_dims=n_dims, n_weights_per_dim=n_weights_per_dim)
    weights = np.empty((n_demos, n_dims * n_weights_per_dim))
    for demo_idx in range(n_demos):
        weights[demo_idx] = promp.weights(Ts[demo_idx], Es[demo_idx])

    return weights, Ts, Es, contexts


plot_training_data = False
n_dims = 12
n_weights_per_dim = 10
# available contexts: "panel_width", "clockwise", "counterclockwise", "left_arm", "right_arm"
context_names = ["panel_width", "clockwise", "counterclockwise"]

#pattern = "data/kuka/20200129_peg_in_hole/csv_processed/*/*.csv"
#pattern = "data/kuka/20191213_carry_heavy_load/csv_processed/*/*.csv"
pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/*/*.csv"

weights, Ts, Es, contexts = generate_training_data(
    pattern, n_weights_per_dim, context_names=context_names, verbose=2)
X = np.hstack((contexts, weights))

random_state = np.random.RandomState(0)
gmm = GMM(n_components=5, random_state=random_state)
gmm.from_samples(X)

n_steps = 100
T_query = np.linspace(0, 1, n_steps)

fig = pv.figure(with_key_callbacks=True)
fig.plot_transform(s=0.1)
tm = UrdfTransformManager()
with open("kuka_lbr/urdf/kuka_lbr.urdf", "r") as f:
    tm.load_urdf(f.read(), mesh_path="kuka_lbr/urdf/")
fig.plot_graph(tm, "kuka_lbr", show_visuals=True)

for panel_width, color, idx in zip([0.3, 0.4, 0.5], ([1.0, 1.0, 0.0], [0.0, 1.0, 1.0], [1.0, 0.0, 1.0]), range(3)):
    print("panel_width = %.2f, color = %s" % (panel_width, color))

    context = np.array([panel_width, 0.0, 1.0])
    conditional_weight_distribution = gmm.condition(np.arange(len(context)), context).to_mvn()
    promp = ProMP(n_dims=n_dims, n_weights_per_dim=n_weights_per_dim)
    promp.from_weight_distribution(
        conditional_weight_distribution.mean,
        conditional_weight_distribution.covariance)

    # mean and covariance in state space
    mean = promp.mean_trajectory(T_query)
    cov = promp.cov_trajectory(T_query).reshape(
        mean.shape[0], mean.shape[1], mean.shape[1], mean.shape[0])

    for t in range(0, len(mean), 1):
        mean_left = pt.transform_from_exponential_coordinates(mean[t, :6])
        mean_right = pt.transform_from_exponential_coordinates(mean[t, 6:12])
        cov_left = cov[t, :6, :6, t]
        cov_right = cov[t, 6:, 6:, t]
        x, y, z = pu.to_projected_ellipsoid(mean_left, cov_left, 0.5, 20)
        surface = Surface(x, y, z, c=color)
        surface.add_artist(fig)
        x, y, z = pu.to_projected_ellipsoid(mean_right, cov_right, 0.5, 20)
        surface = Surface(x, y, z, c=color)

if plot_training_data:
    for E in Es:
        left2base_trajectory = ptr.transforms_from_exponential_coordinates(E[:, :6])
        right2base_trajectory = ptr.transforms_from_exponential_coordinates(E[:, 6:])
        pv.Trajectory(left2base_trajectory, s=0.02).add_artist(fig)
        pv.Trajectory(right2base_trajectory, s=0.02).add_artist(fig)

fig.view_init(azim=0, elev=25)
fig.show()



"""
===============================================
Inverse and Forward Kinematics of a 6-DOF Robot
===============================================

We define a simple trajectory in a robot's workspace, compute the forward
kinematics, then the inverse kinematics and again the forward kinematics
to check the consistency of forward and inverse kinematics.
"""
print(__doc__)


import numpy as np
import matplotlib.pyplot as plt
from pytransform3d.plot_utils import make_3d_axis, Trajectory
from movement_primitives.kinematics import Kinematics


COMPI_URDF = """
<?xml version="1.0"?>
  <robot name="compi">
    <link name="linkmount"/>
    <link name="link1"/>
    <link name="link2"/>
    <link name="link3"/>
    <link name="link4"/>
    <link name="link5"/>
    <link name="link6"/>
    <link name="tcp"/>

    <joint name="joint1" type="revolute">
      <origin xyz="0 0 0" rpy="0 0 0"/>
      <parent link="linkmount"/>
      <child link="link1"/>
      <axis xyz="0 0 1.0"/>
    </joint>

    <joint name="joint2" type="revolute">
      <origin xyz="0 0 0.158" rpy="1.570796 0 0"/>
      <parent link="link1"/>
      <child link="link2"/>
      <axis xyz="0 0 -1.0"/>
    </joint>

    <joint name="joint3" type="revolute">
      <origin xyz="0 0.28 0" rpy="0 0 0"/>
      <parent link="link2"/>
      <child link="link3"/>
      <axis xyz="0 0 -1.0"/>
    </joint>

    <joint name="joint4" type="revolute">
      <origin xyz="0 0 0" rpy="-1.570796 0 0"/>
      <parent link="link3"/>
      <child link="link4"/>
      <axis xyz="0 0 1.0"/>
    </joint>

    <joint name="joint5" type="revolute">
      <origin xyz="0 0 0.34" rpy="1.570796 0 0"/>
      <parent link="link4"/>
      <child link="link5"/>
      <axis xyz="0 0 -1.0"/>
    </joint>

    <joint name="joint6" type="revolute">
      <origin xyz="0 0.346 0" rpy="-1.570796 0 0"/>
      <parent link="link5"/>
      <child link="link6"/>
      <axis xyz="0 0 1.0"/>
    </joint>

    <joint name="jointtcp" type="fixed">
      <origin xyz="0 0 0.05" rpy="0 0 0"/>
      <parent link="link6"/>
      <child link="tcp"/>
    </joint>
  </robot>
"""


kin = Kinematics(COMPI_URDF)
chain = kin.create_chain(["joint%d" % i for i in range(1, 7)], "compi", "tcp", verbose=0)

Q = np.zeros((100, chain.n_joints))
Q[:, 0] = np.linspace(-0.5 * np.pi, 0.5 * np.pi, len(Q))
Q[:, 1] = np.linspace(-0.5 * np.pi, 0.5 * np.pi, len(Q))
Q[:, 2] = np.linspace(-0.5 * np.pi, 0.5 * np.pi, len(Q))
Q[:, 3] = np.linspace(-0.5 * np.pi, 0.5 * np.pi, len(Q))
Q[:, 4] = np.linspace(-0.5 * np.pi, 0.5 * np.pi, len(Q))
Q[:, 5] = np.linspace(-0.5 * np.pi, 0.5 * np.pi, len(Q))

H = chain.forward_trajectory(Q)
random_state = np.random.RandomState(2)
Q2 = chain.inverse_trajectory(H, Q[0], random_state=random_state)
H2 = chain.forward_trajectory(Q2)

plt.figure(figsize=(10, 5))
ax = make_3d_axis(1, 121)
traj = Trajectory(H, s=0.1)
traj.add_trajectory(ax)
ax = make_3d_axis(1, 122)
traj2 = Trajectory(H2, s=0.1)
traj2.add_trajectory(ax)
plt.show()



import numpy as np
import pytransform3d.visualizer as pv
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
import pytransform3d.trajectories as ptr
from movement_primitives.kinematics import Kinematics
from movement_primitives.testing.simulation import get_absolute_path


def animation_callback(step, graph, left_arm, right_arm, left_joint_trajectory, right_joint_trajectory):
    left_arm.forward(left_joint_trajectory[step])
    right_arm.forward(right_joint_trajectory[step])
    graph.set_data()
    return graph


solar_panel_idx = 0
panel_rotation_angle = np.deg2rad(90)
n_steps = 201

q0_left = np.array([-1.57, 0.9, 0, -1.3, 0, 0, -0.55])
q0_right = np.array([1.57, -0.9, 0, 1.3, 0, 0, -0.55])

with open(get_absolute_path("pybullet-only-arms-urdf/urdf/RH5.urdf", "models/robots/rh5_models"), "r") as f:
    kin = Kinematics(f.read(), mesh_path=get_absolute_path("pybullet-only-arms-urdf/urdf/", "models/robots/rh5_models"))
left_arm = kin.create_chain(
    ["ALShoulder1", "ALShoulder2", "ALShoulder3",
     "ALElbow", "ALWristRoll", "ALWristYaw", "ALWristPitch"],
    "RH5_Root_Link", "LTCP_Link")
right_arm = kin.create_chain(
    ["ARShoulder1", "ARShoulder2", "ARShoulder3",
     "ARElbow", "ARWristRoll", "ARWristYaw", "ARWristPitch"],
    "RH5_Root_Link", "RTCP_Link")

left_arm.forward(q0_left)
right_arm.forward(q0_right)

left2base_start = kin.tm.get_transform("ALWristPitch_Link", "RH5_Root_Link")
right2base_start = kin.tm.get_transform("ARWristPitch_Link", "RH5_Root_Link")
start_y = np.hstack((pt.pq_from_transform(left2base_start),
                     pt.pq_from_transform(right2base_start)))

# Confuse the DMP, just for testing purposes...
start_y[3:7] *= -1.0
start_y[10:14] *= -1.0

# Read DMP from your favorite format.
from movement_primitives.io import read_yaml

dmp = read_yaml("rh5_dual_arm_dmp.yaml")
#dmp = read_json("rh5_dual_arm_dmp.json")
#dmp = read_pickle("rh5_dual_arm_dmp.pickle")

# Configure start poses of end effectors. Should be initialized from the
# current end-effector poses.
# start_y is a numpy array with 14 entries: position of the left end effector,
# orientation quaternion of the left end effector, position of the right
# end effector, and orientation quaternion of the right end effector.

# WARNING! There are always two quaternions that represent the exact same
# orientation: q and -q. The problem is that a DMP that moves from q to q
# does not move at all, while a DMP that moves from q to -q moves a lot. We
# have to make sure that start_y always contains the quaternion representation
# that is closest to the previous start_y!
start_y[3:7] = pr.pick_closest_quaternion(start_y[3:7], target_quaternion=dmp.start_y[3:7])
start_y[10:14] = pr.pick_closest_quaternion(start_y[10:14], target_quaternion=dmp.start_y[10:14])
dmp.configure(start_y=start_y)

# Generate trajectory with open loop.
# T - time for each step in the trajectory
# Y_pq - array of shape (number of steps, 14) that contains left position,
#        left orientation, right position, and right orientation for each step
T, Y_pq = dmp.open_loop()

# Generate trajectory with closed loop.
y = np.copy(start_y)  # positions
yd = np.zeros(12)  # velocities
dmp.reset()
dmp.configure(start_y=y)
for i in range(n_steps):
    # get state of end effectors:
    # y = ...
    # yd = ...  # if it is not available, you can ignore this and take the DMP's output

    y, yd = dmp.step(y, yd)  # generate desired poses and velocities

    # here you can send commands to WBC

    Y_pq[i] = y

Y_transforms = np.empty((len(Y_pq), 2, 4, 4))
Y_transforms[:, 0] = ptr.transforms_from_pqs(Y_pq[:, :7])
Y_transforms[:, 1] = ptr.transforms_from_pqs(Y_pq[:, 7:])

lwp_trajectory = Y_transforms[:, 0]
rwp_trajectory = Y_transforms[:, 1]

ltcp2lwp = kin.tm.get_transform("LTCP_Link", "ALWristPitch_Link")
rtcp2rwp = kin.tm.get_transform("RTCP_Link", "ARWristPitch_Link")

left_trajectory = np.array([pt.concat(ltcp2lwp, lwp2base) for lwp2base in lwp_trajectory])
right_trajectory = np.array([pt.concat(rtcp2rwp, rwp2base) for rwp2base in rwp_trajectory])

print("Inverse kinematics...")
random_state = np.random.RandomState(0)
left_joint_trajectory = left_arm.inverse_trajectory(left_trajectory, q0_left, random_state=random_state)
right_joint_trajectory = right_arm.inverse_trajectory(right_trajectory, q0_right, random_state=random_state)

lwp2ltcp = kin.tm.get_transform("ALWristPitch_Link", "LTCP_Link")
rwp2rtcp = kin.tm.get_transform("ARWristPitch_Link", "RTCP_Link")

lwp_trajectory = np.array([pt.concat(lwp2ltcp, ltcp2base) for ltcp2base in left_trajectory])
rwp_trajectory = np.array([pt.concat(rwp2rtcp, rtcp2base) for rtcp2base in right_trajectory])

fig = pv.figure()
fig.plot_transform(s=0.3)

graph = fig.plot_graph(
    kin.tm, "RH5_Root_Link", show_visuals=True, show_collision_objects=False, show_frames=True, s=0.1,
    whitelist=["ALWristPitch_Link", "ARWristPitch_Link", "LTCP_Link", "RTCP_Link"])

left2base_start = kin.tm.get_transform("LTCP_Link", "RH5_Root_Link")
right2base_start = kin.tm.get_transform("RTCP_Link", "RH5_Root_Link")
fig.plot_transform(left2base_start, s=0.15)
fig.plot_transform(right2base_start, s=0.15)

fig.plot_transform(left_trajectory[-1], s=0.15)
fig.plot_transform(right_trajectory[-1], s=0.15)

pv.Trajectory(lwp_trajectory, s=0.05).add_artist(fig)
pv.Trajectory(rwp_trajectory, s=0.05).add_artist(fig)

fig.view_init()
fig.animate(
    animation_callback, len(left_joint_trajectory), loop=True,
    fargs=(graph, left_arm, right_arm, left_joint_trajectory, right_joint_trajectory))
fig.show()



"""Visualization Tools
===================

3D visualization tools for movement primitives.
"""
import numpy as np
import open3d as o3d
import scipy as sp
import pytransform3d.transformations as pt
import pytransform3d.uncertainty as pu


def plot_pointcloud(fig, pcl_points, color, uniform_down_sample=1):
    """Plot point cloud.

    Parameters
    ----------
    fig : pytransform3d.visualizer.Figure
        Figure.

    pcl_points : array-like, shape (n_points, 3)
        Points of the point cloud.

    color : array-like, shape (3,)
        Color as RGB values between 0 and 1.

    uniform_down_sample : int, optional (default: 1)
        Apply downsampling with this factor.

    Returns
    -------
    pcl : open3d.geometry.PointCloud
        Point cloud.
    """
    pcl = o3d.geometry.PointCloud(
        o3d.utility.Vector3dVector(np.array(pcl_points)))
    colors = o3d.utility.Vector3dVector(
        [color for _ in range(len(pcl.points))])
    pcl.colors = colors
    if uniform_down_sample > 1:
        pcl = pcl.uniform_down_sample(uniform_down_sample)
    fig.add_geometry(pcl)
    return pcl


class ToggleGeometry:
    """Key action callback to toggle between showing and hiding a geometry.

    Parameters
    ----------
    fig : pytransform3d.visualizer.Figure
        Figure.

    geometry : open3d.geometry.Geometry
        Geometry.
    """
    def __init__(self, fig, geometry):
        self.fig = fig
        self.geometry = geometry
        self.show = True

    def __call__(self, vis, key, modifier):
        self.show = not self.show
        if self.show and modifier:
            vis.remove_geometry(self.geometry, False)
        elif not self.show and not modifier:
            vis.add_geometry(self.geometry, False)
        return True


def to_ellipsoid(mean, cov):
    """Compute error ellipsoid.

    An error ellipsoid shows equiprobable points.

    Parameters
    ----------
    mean : array-like, shape (3,)
        Mean of distribution

    cov : array-like, shape (3, 3)
        Covariance of distribution

    Returns
    -------
    ellipsoid2origin : array, shape (4, 4)
        Ellipsoid frame in world frame

    radii : array, shape (3,)
        Radii of ellipsoid
    """
    return pu.to_ellipsoid(mean, cov)



"""
=================================
Simulate a DMP in Cartesian Space
=================================

A normal DMP is used to represent a Cartesian trajectory given by positions
and quaternions.
"""
print(__doc__)


import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.dmp import DMP
from movement_primitives.testing.simulation import UR5Simulation


dt = 0.01
execution_time = 1.0

dmp = DMP(n_dims=7, execution_time=execution_time, dt=dt,
          n_weights_per_dim=10, int_dt=0.01, p_gain=2.5)
Y = np.zeros((1001, 7))
T = np.linspace(0, 1, len(Y))
sigmoid = 0.5 * (np.tanh(1.5 * np.pi * (T - 0.5)) + 1.0)
Y[:, 0] = 0.6
Y[:, 1] = -0.2 + 0.4 * sigmoid + np.linspace(-0.4, 0.4, len(Y)) ** 2
Y[:, 2] = 0.45
Y[:, 4] = 1.0
dmp.imitate(T, Y)
dmp.configure(start_y=Y[0], goal_y=Y[-1])

ur5 = UR5Simulation(dt=dt, real_time=False)
for _ in range(4):
    ur5.goto_ee_state(Y[0], wait_time=1.0)
    ur5.stop()

desired_positions, positions, desired_velocities, velocities = \
    ur5.step_through_cartesian(dmp, Y[0], np.zeros(7), 4 * execution_time, closed_loop=True)

P = np.asarray(positions)
dP = np.asarray(desired_positions)
V = np.asarray(velocities)
dV = np.asarray(desired_velocities)

plot_dim = 1
plt.plot(Y[:, plot_dim], label="Demo")
plt.scatter([[0, len(Y)]], [[Y[0, plot_dim], Y[-1, plot_dim]]])
plt.plot(P[:, plot_dim], label="Actual")
plt.scatter([[0, len(P)]], [[P[0, plot_dim], P[-1, plot_dim]]])
plt.plot(dP[:, plot_dim], label="Desired")
plt.scatter([[0, len(dP)]], [[dP[0, plot_dim], dP[-1, plot_dim]]])
T, Y = dmp.open_loop(run_t=1.0)
plt.plot(Y[:, plot_dim], label="Open loop")
plt.scatter([[0, len(Y)]], [[Y[0, plot_dim], Y[-1, plot_dim]]])
plt.legend()
plt.show()



"""
======================
DMP as Potential Field
======================

A Dynamical Movement Primitive defines a potential field that superimposes
several components: transformation system (goal-directed movement), forcing
term (learned shape), and coupling terms (e.g., obstacle avoidance).
"""
print(__doc__)


import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.dmp import DMP, CouplingTermObstacleAvoidance2D
from movement_primitives.dmp_potential_field import plot_potential_field_2d


start_y = np.array([0, 0], dtype=float)
goal_y = np.array([1, 1], dtype=float)
obstacle = np.array([0.85, 0.5])
random_state = np.random.RandomState(1)

dmp = DMP(n_dims=2, n_weights_per_dim=10, dt=0.01, execution_time=1.0)
dmp.configure(start_y=start_y, goal_y=goal_y)

dmp_ft = DMP(n_dims=2, n_weights_per_dim=10, dt=0.01, execution_time=1.0)
dmp_ft.forcing_term.weights_[:, :] = random_state.randn(
    *dmp_ft.forcing_term.weights_.shape) * 500.0
dmp_ft.configure(start_y=start_y, goal_y=goal_y)

dmp_ct = DMP(n_dims=2, n_weights_per_dim=10, dt=0.01, execution_time=1.0)
dmp_ct.forcing_term.weights_[:, :] = dmp_ft.forcing_term.weights_[:, :]
dmp_ct.configure(start_y=start_y, goal_y=goal_y)
coupling_term = CouplingTermObstacleAvoidance2D(obstacle)

n_rows, n_cols = 2, 4
n_subplots = n_rows * n_cols
x_range = -0.2, 1.2
y_range = -0.2, 1.2

position = np.copy(start_y)
velocity = np.zeros_like(start_y)

position_ft = np.copy(start_y)
velocity_ft = np.zeros_like(start_y)

position_ct = np.copy(start_y)
velocity_ct = np.zeros_like(start_y)

plt.figure(figsize=(12, 6))
positions = [position]
positions_ft = [position_ft]
positions_ct = [position_ct]
for i in range(n_subplots):
    ax = plt.subplot(n_rows, n_cols, i + 1, aspect="equal")
    ax.set_title(f"t = {dmp.t:.02f}", backgroundcolor="#ffffffff", y=0.05)

    plot_potential_field_2d(
        ax, dmp_ct, x_range=x_range, y_range=y_range, n_ticks=15,
        obstacle=obstacle)
    plt.plot(start_y[0], start_y[1], "o", color="b", markersize=10)
    plt.plot(goal_y[0], goal_y[1], "o", color="g", markersize=10)
    plt.plot(obstacle[0], obstacle[1], "o", color="y", markersize=10)

    path = np.array(positions)
    plt.plot(path[:, 0], path[:, 1], lw=5, color="g", label="Transformation System")
    path_ft = np.array(positions_ft)
    plt.plot(path_ft[:, 0], path_ft[:, 1], lw=5, color="r", label="+ Forcing Term")
    path_ct = np.array(positions_ct)
    plt.plot(path_ct[:, 0], path_ct[:, 1], lw=5, color="y", label="+ Obstacle Avoidance")

    ax.set_xlim(x_range)
    ax.set_ylim(y_range)
    plt.setp(ax, xticks=(), yticks=())
    if i == 0:
        ax.legend(loc="upper left")

    if i == n_subplots - 1:
        break

    while dmp.t <= dmp.execution_time_ * (1 + i) / (n_subplots - 1):
        position, velocity = dmp.step(position, velocity)
        positions.append(position)
        position_ft, velocity_ft = dmp_ft.step(position_ft, velocity_ft)
        positions_ft.append(position_ft)
        position_ct, velocity_ct = dmp_ct.step(
            position_ct, velocity_ct, coupling_term=coupling_term)
        positions_ct.append(position_ct)
plt.subplots_adjust(
    left=0.0, bottom=0.0, right=1.0, top=1.0, wspace=0.01, hspace=0.01)
plt.show()



import numpy as np
from ..base import PointToPointMovement


class DMPBase(PointToPointMovement):
    """Base class of Dynamical Movement Primitives (DMPs).

    Parameters
    ----------
    n_pos_dims : int
        Number of dimensions of the position that will be controlled.

    n_vel_dims : int
        Number of dimensions of the velocity that will be controlled.
    """
    def __init__(self, n_pos_dims, n_vel_dims):
        super(DMPBase, self).__init__(n_pos_dims, n_vel_dims)

        self.initialized = False

    def reset(self):
        """Reset DMP to initial state and time."""
        self.t = 0.0
        self.last_t = None
        self.current_y = np.copy(self.start_y)
        self.current_yd = np.copy(self.start_yd)


class WeightParametersMixin:
    """Mixin class providing common access methods to forcing term weights.

    This can be used, for instance, for black-box optimization of the weights
    with respect to some cost / objective function in a reinforcement learning
    setting.
    """
    def get_weights(self):
        """Get weight vector of DMP.

        Returns
        -------
        weights : array, shape (N * n_weights_per_dim,)
            Current weights of the DMP. N depends on the type of DMP
        """
        return self.forcing_term.weights_.ravel()

    def set_weights(self, weights):
        """Set weight vector of DMP.

        Parameters
        ----------
        weights : array, shape (N * n_weights_per_dim,)
            New weights of the DMP. N depends on the type of DMP
        """
        self.forcing_term.weights_[:, :] = weights.reshape(*self.forcing_term.shape)

    @property
    def n_weights(self):
        """Total number of weights configuring the forcing term."""
        return np.prod(self.forcing_term.shape)



import numpy as np
import tqdm
from mocap.dataset_loader import load_kuka_dataset, transpose_dataset
from movement_primitives.dmp import DualCartesianDMP
from movement_primitives.dmp_to_state_space_distribution import propagate_to_state_space, estimate_state_distribution
from movement_primitives.visualization import plot_pointcloud, ToggleGeometry
from mocap.cleaning import median_filter
from gmr import GMM
from pytransform3d.batch_rotations import smooth_quaternion_trajectory
import pytransform3d.visualizer as pv
from pytransform3d.urdf import UrdfTransformManager


def generate_training_data(
        pattern, n_weights_per_dim, context_names, smooth_quaterions, verbose=0):
    Ts, Ps, contexts = transpose_dataset(
        load_kuka_dataset(pattern, context_names, verbose=verbose))

    if smooth_quaterions:
        for P in Ps:
            P[:, 3:7] = smooth_quaternion_trajectory(P[:, 3:7])
            P[:, 10:] = smooth_quaternion_trajectory(P[:, 10:])
            P[:, :] = median_filter(P, window_size=5)

    n_demos = len(Ts)

    print("Computing params...")
    params = np.empty((n_demos, 2 * 6 * n_weights_per_dim + 2 * 7 + 2 * 7 + 1))
    exclude_indices = []
    for demo_idx in tqdm.tqdm(range(n_demos)):
        execution_time = Ts[demo_idx][-1]
        dt = np.mean(np.diff(Ts[demo_idx]))
        if dt < 0.005:  # HACK
            exclude_indices.append(demo_idx)
            if verbose:
                tqdm.tqdm.write("Excluding %d, dt = %g" % (demo_idx, dt))
        dmp = DualCartesianDMP(
            execution_time=execution_time, dt=dt,
            n_weights_per_dim=n_weights_per_dim, int_dt=0.01)
        dmp.imitate(Ts[demo_idx], Ps[demo_idx])
        params[demo_idx] = np.hstack((
            dmp.get_weights(), Ps[demo_idx][0], Ps[demo_idx][-1], execution_time))
        if np.any(np.isnan(params[demo_idx])):
            exclude_indices.append(demo_idx)
            if verbose:
                tqdm.tqdm.write("Excluding %d, NaN" % demo_idx)

    params = np.vstack([params[i] for i in range(n_demos) if i not in exclude_indices])
    Ts = [Ts[i] for i in range(n_demos) if i not in exclude_indices]
    Ps = [Ps[i] for i in range(n_demos) if i not in exclude_indices]
    contexts = np.vstack([contexts[i] for i in range(n_demos) if i not in exclude_indices])

    if verbose:
        print("Excluded samples: %s" % (exclude_indices,))

    return params, Ts, Ps, contexts


n_dims = 14
n_weights_per_dim = 10
alpha = 1e-3
kappa = 10.0
# available contexts: "panel_width", "clockwise", "counterclockwise", "left_arm", "right_arm"
context_names = ["panel_width", "clockwise", "counterclockwise"]

#pattern = "data/kuka/20200129_peg_in_hole/csv_processed/*/*.csv"
#pattern = "data/kuka/20191213_carry_heavy_load/csv_processed/*/*.csv"
pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/*/*.csv"
weights, Ts, Ps, contexts = generate_training_data(pattern, n_weights_per_dim, context_names, smooth_quaterions=True, verbose=1)
X = np.hstack((contexts, weights))

random_state = np.random.RandomState(0)
gmm = GMM(n_components=5, random_state=random_state)
gmm.from_samples(X)

n_validation_samples = 100

fig = pv.figure(with_key_callbacks=True)
fig.plot_transform(s=0.1)
tm = UrdfTransformManager()
with open("kuka_lbr/urdf/kuka_lbr.urdf", "r") as f:
    tm.load_urdf(f.read(), mesh_path="kuka_lbr/urdf/")
fig.plot_graph(
    tm, "kuka_lbr", show_frames=True, show_visuals=True,
    whitelist=["kuka_lbr_l_tcp", "kuka_lbr_r_tcp"], s=0.2)


def sample_trajectories(mvn, n_samples, n_dims):
    print("Sampling...")
    sampled_trajectories = mvn.sample(n_samples)
    n_steps = sampled_trajectories.shape[1] // n_dims
    return sampled_trajectories.reshape(n_samples, n_steps, n_dims)


for panel_width, color, idx in zip([0.3, 0.4, 0.5], ([1.0, 1.0, 0.0], [0.0, 1.0, 1.0], [1.0, 0.0, 1.0]), range(3)):
    print("panel_width = %.2f, color = %s" % (panel_width, color))
    context = np.array([panel_width, 0.0, 1.0])
    conditional_weight_distribution = gmm.condition(np.arange(len(context)), context).to_mvn()
    average_execution_time = np.mean([T[-1] for T in Ts])
    trajectories = propagate_to_state_space(conditional_weight_distribution, n_weights_per_dim, average_execution_time,
                                            alpha=alpha, kappa=kappa, dt=0.01, int_dt=0.001, verbose=1)
    conditional_state_distribution = estimate_state_distribution(trajectories, alpha=alpha, kappa=kappa, n_weights_per_dim=n_weights_per_dim)

    mean = conditional_state_distribution.mean.reshape(-1, 14)
    fig.plot_trajectory(mean[:, :7], s=0.05, c=color)
    fig.plot_trajectory(mean[:, 7:], s=0.05, c=color)

    n_samples = 100
    n_dims = 2 * 7
    sampled_trajectories = sample_trajectories(conditional_state_distribution, n_samples, n_dims)

    pcl_points = []
    distances = []
    stds = []
    for P in sampled_trajectories:
        # uncomment to check orientations
        #left = fig.plot_trajectory(P=P[:, :7], s=0.02)
        #right = fig.plot_trajectory(P=P[:, 7:], s=0.02)
        pcl_points.extend(P[:, :3])
        pcl_points.extend(P[:, 7:10])

        ee_distances = np.linalg.norm(P[:, :3] - P[:, 7:10], axis=1)
        distances.append(np.mean(ee_distances))
        stds.append(np.std(ee_distances))
    print("Mean average distance of end-effectors = %.2f, mean std. dev. = %.3f"
          % (np.mean(distances), np.mean(stds)))

    pcl = plot_pointcloud(fig, pcl_points, color)
    key = ord(str((idx + 1) % 10))
    fig.visualizer.register_key_action_callback(key, ToggleGeometry(fig, pcl))

fig.show()



import math
import numpy as np
try:
    from ..dmp_fast import phase
except ImportError:
    from ._canonical_system import phase


class ForcingTerm:
    """Defines the shape of a DMP.

    Parameters
    ----------
    n_dims : int
        State space dimensions.

    n_weights_per_dim : int, optional (default: 10)
        Number of weights of the function approximator per dimension.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    overlap : float
        Value at which radial basis functions overlap.

    alpha_z : float
        Value of the alpha parameter of the canonical system.

    Attributes
    ----------
    weights_ : array, shape (n_dims, n_weights_per_dim)
        Weights of the forcing term.
    """
    def __init__(self, n_dims, n_weights_per_dim, goal_t, start_t, overlap,
                 alpha_z):
        if n_weights_per_dim <= 1:
            raise ValueError("The number of weights per dimension must be > 1!")
        self.n_weights_per_dim = n_weights_per_dim
        if start_t >= goal_t:
            raise ValueError("Goal must be chronologically after start!")
        self.goal_t = goal_t
        self.start_t = start_t
        self.overlap = overlap
        self.alpha_z = alpha_z

        self._init_rbfs(n_dims, n_weights_per_dim, start_t)

    def _init_rbfs(self, n_dims, n_weights_per_dim, start_t):
        self.log_overlap = float(-math.log(self.overlap))
        self.execution_time = self.goal_t - self.start_t
        self.weights_ = np.zeros((n_dims, n_weights_per_dim))
        self.centers = np.empty(n_weights_per_dim)
        self.widths = np.empty(n_weights_per_dim)
        # -1 because we want the last entry to be execution_time
        step = self.execution_time / (self.n_weights_per_dim - 1)
        # do first iteration outside loop because we need access to i and i - 1 in loop
        t = start_t
        self.centers[0] = phase(t, self.alpha_z, self.goal_t, self.start_t)
        for i in range(1, self.n_weights_per_dim):
            # normally lower_border + i * step but lower_border is 0
            t = i * step
            self.centers[i] = phase(t, self.alpha_z, self.goal_t, self.start_t)
            # Choose width of RBF basis functions automatically so that the
            # RBF centered at one center has value overlap at the next center
            diff = self.centers[i] - self.centers[i - 1]
            self.widths[i - 1] = self.log_overlap / diff ** 2
        # Width of last Gaussian cannot be calculated, just use the same width
        # as the one before
        self.widths[self.n_weights_per_dim - 1] = self.widths[
            self.n_weights_per_dim - 2]

    def _activations(self, z):
        z = np.atleast_2d(z)  # 1 x n_steps
        squared_dist = (z - self.centers[:, np.newaxis]) ** 2
        activations = np.exp(-self.widths[:, np.newaxis] * squared_dist)
        activations /= activations.sum(axis=0)  # normalize
        return activations

    def design_matrix(self, T):  # returns: n_weights_per_dim x n_steps
        Z = phase(T, alpha=self.alpha_z, goal_t=T[-1], start_t=T[0])
        return Z[np.newaxis, :] * self._activations(Z)

    def phase(self, t):
        return phase(t, alpha=self.alpha_z, goal_t=self.goal_t,
                     start_t=self.start_t)

    def forcing_term(self, z):
        z = np.atleast_1d(z)
        activations = self._activations(z)
        return z[np.newaxis, :] * self.weights_.dot(activations)

    def __call__(self, t):
        return self.forcing_term(self.phase(t))

    @property
    def shape(self):
        """Shape (n_dims, n_weights_per_dim) of weights configuring the forcing term."""
        return self.weights_.shape



"""
===================
Two Coupled 1D DMPs
===================

Two 1D DMPs are spatially coupled with a virtual spring that forces them to
keep a distance. One of them is the leader DMP and the other one is the
follower DMP.
"""
print(__doc__)


import matplotlib.pyplot as plt
import numpy as np
from movement_primitives.dmp import DMP, CouplingTermPos1DToPos1D


dt = 0.01
execution_time = 2.0
desired_distance = 0.5
dmp = DMP(n_dims=2, execution_time=execution_time, dt=dt, n_weights_per_dim=200)
coupling_term = CouplingTermPos1DToPos1D(
    desired_distance=desired_distance, lf=(1.0, 0.0), k=1.0)

T = np.linspace(0.0, execution_time, 101)
Y = np.empty((len(T), 2))
Y[:, 0] = np.cos(2.5 * np.pi * T)
Y[:, 1] = 0.5 + np.cos(1.5 * np.pi * T)
dmp.imitate(T, Y)

fig = plt.figure(figsize=(10, 5))

ax1 = fig.add_subplot(131)
ax1.set_title("Dimension 1")
ax1.set_ylim((-3, 3))
ax2 = fig.add_subplot(132)
ax2.set_title("Dimension 2")
ax2.set_ylim((-3, 3))
ax1.plot(T, Y[:, 0], label="Demo")
ax1.scatter([T[0], T[-1]], [Y[0, 0], Y[-1, 0]])
ax2.plot(T, Y[:, 1], label="Demo")
ax2.scatter([T[0], T[-1]], [Y[0, 1], Y[-1, 1]])

dmp.configure(start_y=Y[0], goal_y=Y[-1])
T, Y = dmp.open_loop()
ax1.plot(T, Y[:, 0], label="Reproduction")
ax1.scatter([T[0], T[-1]], [Y[0, 0], Y[-1, 0]])
ax2.plot(T, Y[:, 1], label="Reproduction")
ax2.scatter([T[0], T[-1]], [Y[0, 1], Y[-1, 1]])

dmp.configure(start_y=Y[0], goal_y=Y[-1])
T, Y = dmp.open_loop(coupling_term=coupling_term)
ax1.plot(T, Y[:, 0], label="Coupled 1")
ax2.plot(T, Y[:, 1], label="Coupled 2")
ax1.scatter([T[0], T[-1]], [Y[0, 0], Y[-1, 0]])
ax2.scatter([T[0], T[-1]], [Y[0, 1], Y[-1, 1]])

ax1.legend(loc="best")

ax = fig.add_subplot(133)
ax.set_title("Distance")
ax.set_ylim((-3, 3))
ax.plot(T, Y[:, 1] - Y[:, 0], label="Actual distance")
ax.plot([T[0], T[-1]], [desired_distance, desired_distance],
        label="Desired distance")

ax.legend(loc="best")

plt.show()



"""
=========================
Stepwise Execution of DMP
=========================

Since DMPs are executed open loop in most examples, in this case we will
demonstrate how to perform stepwise execution of a DMP, which is necessary to
incorporate sensor data in the execution.
"""
import time
import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.dmp import DMP, CouplingTermObstacleAvoidance2D


execution_time = 1.0
dt = 0.001
start_y = np.zeros(2)
goal_y = np.ones(2)

dmp = DMP(n_dims=2, execution_time=execution_time, dt=dt, n_weights_per_dim=3)
dmp.configure(start_y=start_y, goal_y=goal_y)
dmp.set_weights(np.array([-50.0, 100.0, 300.0, -200.0, -200.0, -200.0]))
obstacle_position = np.array([0.92, 0.5])
coupling_term = CouplingTermObstacleAvoidance2D(obstacle_position, fast=True)

# on a real system, these would be measured values
y = np.copy(dmp.start_y)
yd = np.copy(dmp.start_yd)

T = []
Y = []

total_time = 0.0
for t in np.arange(0.0, 2.0, dt):  # on real system, t would be measured time
    T.append(t)
    Y.append(y)
    start_time = time.time()
    y, yd = dmp.step(y, yd, coupling_term, step_function="rk4-cython")
    total_time += time.time() - start_time
T.append(2.0)
Y.append(y)

T = np.array(T)
Y = np.array(Y)

print("Total runtime of DMP: %g s" % total_time)

fig = plt.figure()
ax = fig.add_subplot(111)
ax.set_xlabel("x")
ax.set_ylabel("y")
ax.plot(Y[:, 0], Y[:, 1], label="Obstacle avoidance")
ax.scatter(start_y[0], start_y[1], c="r", label="Start")
ax.scatter(goal_y[0], goal_y[1], c="g", label="Goal")
ax.scatter(obstacle_position[0], obstacle_position[1], c="y", label="Obstacle")
ax.legend()
plt.tight_layout()
plt.show()



"""
=====================
Velocities of a ProMP
=====================

ProMPs can generate positions and velocities of a trajectory.
"""
print(__doc__)


import numpy as np
from movement_primitives.promp import ProMP
import matplotlib.pyplot as plt


n_weights_per_dim = 100

promp = ProMP(n_dims=1, n_weights_per_dim=n_weights_per_dim)

random_state = np.random.RandomState(10)
n_demos = 10
n_steps = 101
T = np.empty((n_demos, n_steps))
T[:, :] = np.linspace(0.0, 1.0, n_steps)
Y = np.empty((n_demos, n_steps, 1))
for demo_idx in range(n_demos):
    Y[demo_idx] = np.cos(2 * np.pi * T[demo_idx] + random_state.randn() * 0.1)[:, np.newaxis]
    Y[demo_idx, :, 0] += random_state.randn(n_steps) * 0.01
promp.imitate(T, Y, verbose=1)

random_state = np.random.RandomState(0)
samples = promp.sample_trajectories(T[0], 10, random_state)

mean_trajectory = promp.mean_trajectory(T[0])
std_trajectory = np.sqrt(promp.var_trajectory(T[0]))

mean_velocities = promp.mean_velocities(T[0])
std_velocities = np.sqrt(promp.var_velocities(T[0]))

dt = T[0, 1] - T[0, 0]
int_velocities = mean_trajectory[0] + np.cumsum(mean_velocities * dt)

fig = plt.figure(figsize=(10, 5))

ax = fig.add_subplot(121)
for demo_idx in range(n_demos):
    label = "Demo" if demo_idx == 0 else None
    ax.plot(T[demo_idx, :], Y[demo_idx, :], c="k", alpha=0.1, label=label)
ax.plot(T[0], mean_trajectory, label="Reproduction", c="r", lw=3)
ax.plot(T[0], int_velocities, label="Integrated velocities", c="b")
factor = 2
ax.fill_between(
    T[0],
    mean_trajectory[:, 0] - factor * std_trajectory[:, 0],
    mean_trajectory[:, 0] + factor * std_trajectory[:, 0],
    alpha=0.3, label="1-$\sigma$ interval")
for sample in samples:
    ax.plot(T[0], sample, c="g", alpha=0.3)
ax.legend(loc="best")

ax = fig.add_subplot(122)
ax.plot(T[0], mean_velocities, label="Velocities", c="r", lw=3)
factor = 2
ax.fill_between(
    T[0],
    mean_velocities[:, 0] - factor * std_velocities[:, 0],
    mean_velocities[:, 0] + factor * std_velocities[:, 0],
    alpha=0.3, label="1-$\sigma$ interval")
for i, sample in enumerate(samples):
    label = "Samples" if i == 0 else None
    ax.plot(T[0], np.gradient(sample, axis=0) * n_steps, c="g", alpha=0.3,
            label=label)
ax.legend(loc="best")

plt.show()



"""
===================================
Pose Coupling of Dual Cartesian DMP
===================================

A dual Cartesian DMP is learned from an artificially generated demonstration
and replayed with and without a coupling of the pose of the two end effectors.

The red line indicates the DMP without coupling term and the orange line marks
the DMP with coupling term.
"""
print(__doc__)


import warnings
import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.dmp import DualCartesianDMP, CouplingTermDualCartesianPose
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
import pytransform3d.trajectories as ptr
from pytransform3d.urdf import UrdfTransformManager
from movement_primitives.testing.simulation import SimulationMockup


dt = 0.01
int_dt = 0.001
execution_time = 1.0

desired_distance = np.array([  # right arm to left arm
    [1.0, 0.0, 0.0, 0.0],
    [0.0, 1.0, 0.0, 0.0],
    [0.0, 0.0, 1.0, -1.2],
    [0.0, 0.0, 0.0, 1.0]
])
desired_distance[:3, :3] = pr.matrix_from_compact_axis_angle([np.deg2rad(180), 0, 0])
ct = CouplingTermDualCartesianPose(
    desired_distance=desired_distance, couple_position=True, couple_orientation=True,
    lf=(1.0, 0.0), k=1, c1=0.1, c2=1000)  # c2=10000 in simulation

rh5 = SimulationMockup(dt=dt)

Y = np.zeros((1001, 14))
T = np.linspace(0, 1, len(Y))
sigmoid = 0.5 * (np.tanh(1.5 * np.pi * (T - 0.5)) + 1.0)
radius = 0.5

circle1 = radius * np.cos(np.deg2rad(90) + np.deg2rad(90) * sigmoid)
circle2 = radius * np.sin(np.deg2rad(90) + np.deg2rad(90) * sigmoid)
Y[:, 0] = circle1
Y[:, 1] = 0.55
Y[:, 2] = circle2
R_three_fingers_front = pr.matrix_from_axis_angle([0, 0, 1, 0.5 * np.pi])
R_to_center_start = pr.matrix_from_axis_angle([1, 0, 0, np.deg2rad(0)])
# introduces coupling error (default goal: -90; error at: -110)
R_to_center_end = pr.matrix_from_axis_angle([1, 0, 0, np.deg2rad(-110)])
q_start = pr.quaternion_from_matrix(R_three_fingers_front.dot(R_to_center_start))
q_end = -pr.quaternion_from_matrix(R_three_fingers_front.dot(R_to_center_end))
for i, t in enumerate(T):
    Y[i, 3:7] = pr.quaternion_slerp(q_start, q_end, t)

circle1 = radius * np.cos(np.deg2rad(270) + np.deg2rad(90) * sigmoid)
circle2 = radius * np.sin(np.deg2rad(270) + np.deg2rad(90) * sigmoid)
Y[:, 7] = circle1
Y[:, 8] = 0.55
Y[:, 9] = circle2
R_three_fingers_front = pr.matrix_from_axis_angle([0, 0, 1, 0.5 * np.pi])
R_to_center_start = pr.matrix_from_axis_angle([1, 0, 0, np.deg2rad(-180)])
R_to_center_end = pr.matrix_from_axis_angle([1, 0, 0, np.deg2rad(-270)])
q_start = pr.quaternion_from_matrix(R_three_fingers_front.dot(R_to_center_start))
q_end = pr.quaternion_from_matrix(R_three_fingers_front.dot(R_to_center_end))
for i, t in enumerate(T):
    Y[i, 10:] = pr.quaternion_slerp(q_start, q_end, t)

tm = UrdfTransformManager()
try:
    with open("abstract-urdf-gripper/urdf/GripperLeft.urdf", "r") as f:
        tm.load_urdf(f.read(), mesh_path="abstract-urdf-gripper/urdf/")
    with open("abstract-urdf-gripper/urdf/GripperRight.urdf", "r") as f:
        tm.load_urdf(f.read(), mesh_path="abstract-urdf-gripper/urdf/")
except FileNotFoundError:
    warnings.warn("URDF not found")

tm.add_transform("ALWristPitch_Link", "base", np.eye(4))
tm.add_transform("ARWristPitch_Link", "base", np.eye(4))

dmp = DualCartesianDMP(
    execution_time=execution_time, dt=dt,
    n_weights_per_dim=10, int_dt=int_dt, p_gain=0.0)
dmp.imitate(T, Y)
ax = pr.plot_basis(ax_s=0.8, s=0.1)
for coupling_term, color in [(ct, "orange"), (None, "red")]:
    dmp.reset()

    rh5.goto_ee_state(Y[0])
    desired_positions, positions, desired_velocities, velocities = \
        rh5.step_through_cartesian(dmp, Y[0], np.zeros(12), execution_time,
                                   coupling_term=coupling_term)
    P = np.asarray(positions)
    V = np.asarray(velocities)

    ptr.plot_trajectory(ax=ax, P=P[:, :7], s=0.05, color=color, show_direction=False)
    ptr.plot_trajectory(ax=ax, P=P[:, 7:], s=0.05, color=color, show_direction=False)
    for t in range(0, len(P), 50):
        gripper_left2base = pt.transform_from_pq(P[t, :7])
        gripper_right2base = pt.transform_from_pq(P[t, 7:])
        tm.add_transform("ALWristPitch_Link", "base", gripper_left2base)
        tm.add_transform("ARWristPitch_Link", "base", gripper_right2base)
        ax = tm.plot_visuals(frame="base", ax=ax)
    ax.view_init(elev=0, azim=90)
plt.show()



"""
========================
Obstacle Avoidance in 3D
========================

Plots a DMP in 3D that goes through a point obstacle without obstacle
avoidance. Then we start use the same DMP to start from multiple random
start positions with an activated coupling term for obstacle avoidance.
"""
print(__doc__)


import matplotlib.pyplot as plt
import numpy as np
import pytransform3d.plot_utils as ppu
from movement_primitives.dmp import DMP, CouplingTermObstacleAvoidance3D


execution_time = 1.0
start_y = np.zeros(3)
goal_y = np.ones(3)
obstacle = 0.5 * np.ones(3)
random_state = np.random.RandomState(42)

dmp = DMP(n_dims=len(start_y), execution_time=execution_time,
          n_weights_per_dim=10)
dmp.configure(start_y=start_y, goal_y=goal_y)

ax = ppu.make_3d_axis(1)
T, Y = dmp.open_loop()
ax.plot(Y[:, 0], Y[:, 1], Y[:, 2], c="g")
ax.scatter(start_y[0], start_y[1], start_y[2], c="r")
ax.scatter(goal_y[0], goal_y[1], goal_y[2], c="g")
ax.scatter(obstacle[0], obstacle[1], obstacle[2], c="y")
for _ in range(20):
    start_y_random = 0.2 + 0.2 * random_state.randn(3)
    coupling_term = CouplingTermObstacleAvoidance3D(obstacle)
    dmp.configure(start_y=start_y_random)
    T, Y = dmp.open_loop(coupling_term=coupling_term)
    ax.plot(Y[:, 0], Y[:, 1], Y[:, 2], c="b")
    ax.scatter(start_y_random[0], start_y_random[1], start_y_random[2], c="b")
ax.set_xlim((-0.2, 1.2))
ax.set_ylim((-0.2, 1.2))
ax.set_zlim((-0.2, 1.2))
plt.show()



"""
=======================
DMP with Final Velocity
=======================

Not all DMPs allow a final velocity > 0. In this case we analyze the effect
of changing final velocities in an appropriate variation of the DMP
formulation that allows to set the final velocity.
"""
print(__doc__)


import matplotlib.pyplot as plt
import numpy as np
from movement_primitives.dmp import DMPWithFinalVelocity


dt = 0.01
execution_time = 1.0
T = np.arange(0, execution_time + dt, dt)
Y = np.column_stack((np.cos(np.pi * T), -np.cos(np.pi * T)))

dmp = DMPWithFinalVelocity(n_dims=2, execution_time=execution_time)
dmp.imitate(T, Y)

plt.figure(figsize=(10, 8))
ax1 = plt.subplot(221)
ax1.set_title("Dimension 1")
ax1.set_xlabel("Time")
ax1.set_ylabel("Position")
ax2 = plt.subplot(222)
ax2.set_title("Dimension 2")
ax2.set_xlabel("Time")
ax2.set_ylabel("Position")
ax3 = plt.subplot(223)
ax3.set_xlabel("Time")
ax3.set_ylabel("Velocity")
ax4 = plt.subplot(224)
ax4.set_xlabel("Time")
ax4.set_ylabel("Velocity")
ax1.plot(T, Y[:, 0], label="Demo")
ax2.plot(T, Y[:, 1], label="Demo")
ax3.plot(T, np.gradient(Y[:, 0]) / dmp.dt_)
ax4.plot(T, np.gradient(Y[:, 1]) / dmp.dt_)
ax3.scatter([T[-1]], (Y[-1, 0] - Y[-2, 0]) / dmp.dt_)
ax4.scatter([T[-1]], (Y[-1, 1] - Y[-2, 1]) / dmp.dt_)
for goal_yd in [0.0, 1.0, 2.0]:
    dmp.configure(goal_yd=np.array([goal_yd, goal_yd]))
    T, Y = dmp.open_loop(run_t=execution_time)
    ax1.plot(T, Y[:, 0], label="goal_yd = %g" % goal_yd)
    ax2.plot(T, Y[:, 1], label="goal_yd = %g" % goal_yd)
    ax3.plot(T, np.gradient(Y[:, 0]) / dmp.dt_)
    ax4.plot(T, np.gradient(Y[:, 1]) / dmp.dt_)
    ax3.scatter([T[-1]], [goal_yd])
    ax4.scatter([T[-1]], [goal_yd])
ax1.legend()
plt.tight_layout()
plt.show()



"""
===================
Two Coupled 3D DMPs
===================

Two 3D DMPs are spatially coupled with a virtual spring that forces them to
keep a distance. One of them is the leader DMP and the other one is the
follower DMP.
"""
print(__doc__)


import matplotlib.pyplot as plt
import numpy as np
from movement_primitives.dmp import DMP, CouplingTermPos3DToPos3D


dt = 0.01

dmp = DMP(n_dims=6, execution_time=1.0, dt=dt, n_weights_per_dim=10,
          int_dt=0.0001)
ct = CouplingTermPos3DToPos3D(desired_distance=np.array([0.1, 0.5, 1.0]),
                              lf=(0.0, 1.0), k=1.0, c1=30.0, c2=100.0)

T = np.linspace(0.0, 1.0, 101)
Y = np.empty((len(T), 6))
Y[:, 0] = np.cos(np.pi * T)
Y[:, 1] = np.sin(np.pi * T)
Y[:, 2] = np.sin(2 * np.pi * T)
Y[:, 3] = np.cos(np.pi * T)
Y[:, 4] = np.sin(np.pi * T)
Y[:, 5] = 0.5 + np.sin(2 * np.pi * T)
dmp.imitate(T, Y)

fig = plt.figure(figsize=(10, 5))
gs = fig.add_gridspec(3, 2)
ax = fig.add_subplot(gs[:, 0], projection="3d")

ax.plot(Y[:, 0], Y[:, 1], Y[:, 2], label="Demo 1")
ax.plot(Y[:, 3], Y[:, 4], Y[:, 5], label="Demo 2")

dmp.configure(start_y=Y[0], goal_y=Y[-1])
T, Y = dmp.open_loop()
ax.plot(Y[:, 0], Y[:, 1], Y[:, 2], label="Reproduction 1")
ax.plot(Y[:, 3], Y[:, 4], Y[:, 5], label="Reproduction 2")

dmp.configure(start_y=Y[0], goal_y=Y[-1])
T, Y = dmp.open_loop(coupling_term=ct)
ax.plot(Y[:, 0], Y[:, 1], Y[:, 2], label="Coupled 1")
ax.plot(Y[:, 3], Y[:, 4], Y[:, 5], label="Coupled 2")

ax.legend(loc="best")

ax = fig.add_subplot(gs[0, 1])
ax.plot(T, Y[:, 3] - Y[:, 0], label="Actual distance (X)")
ax.plot([T[0], T[-1]], [ct.desired_distance[0], ct.desired_distance[0]],
        label="Desired distance (X)")

ax = fig.add_subplot(gs[1, 1])
ax.plot(T, Y[:, 4] - Y[:, 1], label="Actual distance (Y)")
ax.plot([T[0], T[-1]], [ct.desired_distance[1], ct.desired_distance[1]],
        label="Desired distance (Y)")

ax = fig.add_subplot(gs[2, 1])
ax.plot(T, Y[:, 5] - Y[:, 2], label="Actual distance (Z)")
ax.plot([T[0], T[-1]], [ct.desired_distance[2], ct.desired_distance[2]],
        label="Desired distance (Z)")

plt.show()



import numpy as np


def check_1d_array_length(var, var_name, expected_length):
    """Check length of 1D array.

    Parameters
    ----------
    var : array-like
        1D array to be checked.

    var_name : str
        Name of the variable. To be used in the error message.

    expected_length : int
        Expected length of the array.

    Raises
    ------
    ValueError
        If the length of the configured meta parameter is not correct.
    """
    actual_length = len(var)
    if actual_length != expected_length:
        s = "s" if expected_length > 1 else ""
        raise ValueError(
            f"Expected {var_name} with {expected_length} element{s}, "
            f"got {actual_length}.")


def ensure_1d_array(value, n_dims, var_name):
    """Process scalar or array-like input to ensure it is a 1D numpy array.

    Parameters
    ----------
    value : float or array-like, shape (n_dims,)
        Argument to be processed.

    n_dims : int
        Expected length of the 1d array.

    var_name : str
        Name of the variable in case an exception has to be raised.

    Returns
    -------
    value : array, shape (n_dims,)
        1D numpy array with dtype float.

    Raises
    ------
    ValueError
        If the argument is not compatible.
    """
    value = np.atleast_1d(value).astype(float)
    if value.ndim == 1 and value.shape[0] == 1:
        value = np.repeat(value, n_dims)
    if value.ndim > 1 or value.shape[0] != n_dims:
        raise ValueError(
            f"{var_name} has incorrect shape, expected ({n_dims},) "
            f"got {value.shape}")
    return value



import numpy as np
import matplotlib.pyplot as plt
import tqdm
from pytransform3d import trajectories as ptr
from pytransform3d import transformations as pt
from pytransform3d.batch_rotations import smooth_quaternion_trajectory
from mocap.dataset_loader import load_kuka_dataset, transpose_dataset, smooth_dual_arm_trajectories_pq
from mocap.cleaning import median_filter
from gmr import GMM

from movement_primitives.promp import ProMP
from movement_primitives.plot import plot_distribution_in_rows, PALETTE


def generate_training_data(
        pattern, n_weights_per_dim, context_names, smooth_quaterions, verbose=0):
    Ts, Ps, contexts = transpose_dataset(
        load_kuka_dataset(pattern, context_names, verbose=verbose))

    if smooth_quaterions:
        smooth_dual_arm_trajectories_pq(Ps)

    Ps_left = []
    Ps_right = []
    Ps_diff = []
    print("Computing relative poses...")
    for P in tqdm.tqdm(Ps):
        P_left = P[:, :7]
        P_right = P[:, 7:]

        left2base = ptr.transforms_from_pqs(P_left)
        right2base = ptr.transforms_from_pqs(P_right)

        left2right = np.empty_like(left2base)
        for t in range(len(left2right)):
            left2right[t] = pt.concat(left2base[t], pt.invert_transform(right2base[t], check=False), check=False)
        P_diff = ptr.pqs_from_transforms(left2right)

        if smooth_quaterions:
            P_diff[:, 3:] = smooth_quaternion_trajectory(P_diff[:, 3:])
            P_diff[:, :] = median_filter(P_diff, window_size=5)

        Ps_left.append(P_left)
        Ps_right.append(P_right)
        Ps_diff.append(P_diff)

    n_demos = len(Ts)
    n_dims = Ps_left[0].shape[1]

    promp = ProMP(n_dims=n_dims, n_weights_per_dim=n_weights_per_dim)
    weights_left = np.empty((n_demos, n_dims * n_weights_per_dim))
    weights_right = np.empty((n_demos, n_dims * n_weights_per_dim))
    weights_diff = np.empty((n_demos, n_dims * n_weights_per_dim))
    for demo_idx in range(n_demos):
        weights_left[demo_idx] = promp.weights(Ts[demo_idx], Ps_left[demo_idx])
        weights_right[demo_idx] = promp.weights(Ts[demo_idx], Ps_right[demo_idx])
        weights_diff[demo_idx] = promp.weights(Ts[demo_idx], Ps_diff[demo_idx])

    return weights_left, weights_right, weights_diff, Ts, Ps_left, Ps_right, Ps_diff, contexts


plot_training_data = False
n_dims = 7
n_weights_per_dim = 10
random_state = np.random.RandomState(0)
# available contexts: "panel_width", "clockwise", "counterclockwise", "left_arm", "right_arm"
context_names = ["panel_width", "clockwise", "counterclockwise"]

pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/*/*.csv"

weights_left, weights_right, weights_diff, Ts, Ps_left, Ps_right, Ps_diff, contexts = generate_training_data(
    pattern, n_weights_per_dim, context_names=context_names,
    smooth_quaterions=True, verbose=2)

gmm_left = GMM(n_components=3, random_state=random_state)
gmm_left.from_samples(np.hstack((contexts, weights_left)))
gmm_right = GMM(n_components=3, random_state=random_state)
gmm_right.from_samples(np.hstack((contexts, weights_right)))
gmm_diff = GMM(n_components=3, random_state=random_state)
gmm_diff.from_samples(np.hstack((contexts, weights_diff)))

n_validation_samples = 100
n_steps = 100
T_query = np.linspace(0, 1, n_steps)
context_indices = np.arange(len(context_names))

plt.figure()
axes = None
for panel_width, color, idx in zip([0.3, 0.4, 0.5], PALETTE[:3], range(3)):
    print("panel_width = %.2f, color = %s" % (panel_width, color))
    context = np.array([panel_width, 0.0, 1.0])

    cgmm_left = gmm_left.condition(context_indices, context).to_mvn()
    cpromp_left = ProMP(n_dims=n_dims, n_weights_per_dim=n_weights_per_dim)
    cpromp_left.from_weight_distribution(cgmm_left.mean, cgmm_left.covariance)

    cgmm_right = gmm_right.condition(context_indices, context).to_mvn()
    cpromp_right = ProMP(n_dims=n_dims, n_weights_per_dim=n_weights_per_dim)
    cpromp_right.from_weight_distribution(cgmm_right.mean, cgmm_right.covariance)

    cgmm_diff = gmm_diff.condition(context_indices, context).to_mvn()
    cpromp_diff = ProMP(n_dims=n_dims, n_weights_per_dim=n_weights_per_dim)
    cpromp_diff.from_weight_distribution(cgmm_diff.mean, cgmm_diff.covariance)

    # mean and standard deviation in state space
    mean_left = cpromp_left.mean_trajectory(T_query)
    std_left = np.sqrt(cpromp_left.var_trajectory(T_query))

    mean_right = cpromp_right.mean_trajectory(T_query)
    std_right = np.sqrt(cpromp_right.var_trajectory(T_query))

    mean_diff = cpromp_diff.mean_trajectory(T_query)
    std_diff = np.sqrt(cpromp_diff.var_trajectory(T_query))

    mean = np.hstack((mean_left, mean_right, mean_diff))
    std = np.hstack((std_left, std_right, std_diff))

    axes = plot_distribution_in_rows(
        mean, std, T_query, axes=axes, subplot_shape=(7, 3),
        color=color, transpose=True, alpha=0.3, fill_between=True,
        std_factors=[2], label="panel width = %.2f" % panel_width)

    for i in range(mean.shape[1]):
        m = np.mean(mean[:, i])
        axes[i].set_ylim((m - 0.25, m + 0.25))

    axes[-1].legend(loc="upper left")

plt.show()



"""
=============================
Simulate Spring-Damper System
=============================

A spring-damper system is used to control the Cartesian pose of a robot arm.
"""
print(__doc__)


import numpy as np
import matplotlib.pyplot as plt
from movement_primitives.spring_damper import SpringDamper
from movement_primitives.testing.simulation import UR5Simulation


dt = 0.01

sd = SpringDamper(n_dims=7, k=2.0, c=None, dt=dt, int_dt=0.001)
start = np.array([0.6, 0.2, 0.45, 1.0, 0.0, 0.0, 0.0])
goal = np.array([0.6, 0.0, 0.45, 1.0, 0.0, 0.0, 0.0])
sd.configure(start_y=start, goal_y=goal)

ur5 = UR5Simulation(dt=dt, real_time=False)
for _ in range(4):
    ur5.goto_ee_state(start, wait_time=1.0)
    ur5.stop()

desired_positions, positions, desired_velocities, velocities = \
    ur5.step_through_cartesian(sd, start, np.zeros(7), 10.0, closed_loop=True)

P = np.asarray(positions)
dP = np.asarray(desired_positions)
V = np.asarray(velocities)
dV = np.asarray(desired_velocities)

plot_dim = 1
plt.figure(figsize=(10, 5))
plt.subplot(131)
plt.title("Position")
T, Y = sd.open_loop(run_t=10.0)
plt.plot(P[:, plot_dim], label="Actual")
plt.scatter([[0, len(P)]], [[P[0, plot_dim], P[-1, plot_dim]]])
plt.plot(dP[:, plot_dim], label="Desired")
plt.scatter([[0, len(dP)]], [[dP[0, plot_dim], dP[-1, plot_dim]]])
plt.plot(Y[:, plot_dim], label="Open loop")
plt.scatter([[0, len(Y)]], [[Y[0, plot_dim], Y[-1, plot_dim]]])
plt.legend()
plt.subplot(132)
plt.title("Velocity")
plt.plot(np.gradient(P[:, plot_dim]) / dt, label="Actual")
plt.plot(np.gradient(dP[:, plot_dim]) / dt, label="Desired")
plt.plot(np.gradient(Y[:, plot_dim]) / dt, label="Open loop")
plt.subplot(133)
plt.title("Acceleration")
plt.plot(np.gradient(np.gradient(P[:, plot_dim])) / dt ** 2, label="Actual")
plt.plot(np.gradient(np.gradient(dP[:, plot_dim])) / dt ** 2, label="Desired")
plt.plot(np.gradient(np.gradient(Y[:, plot_dim])) / dt ** 2, label="Open loop")
plt.tight_layout()
plt.show()



import numpy as np
import pytransform3d.rotations as pr
from ..utils import ensure_1d_array
from ._base import DMPBase
from ._forcing_term import ForcingTerm
from ._canonical_system import canonical_system_alpha
from ._dmp import (dmp_open_loop, dmp_imitate, ridge_regression,
                   DMP_STEP_FUNCTIONS, DEFAULT_DMP_STEP_FUNCTION, phase)

def dmp_step_quaternion_python(
        last_t, t,
        current_y, current_yd,
        goal_y, goal_yd, goal_ydd,
        start_y, start_yd, start_ydd,
        goal_t, start_t, alpha_y, beta_y,
        forcing_term,
        coupling_term=None,
        coupling_term_precomputed=None,
        int_dt=0.001,
        smooth_scaling=False):
    """Integrate quaternion DMP for one step with Euler integration.

    Parameters
    ----------
    last_t : float
        Time at last step.

    t : float
        Time at current step.

    current_y : array, shape (7,)
        Current position. Will be modified.

    current_yd : array, shape (6,)
        Current velocity. Will be modified.

    goal_y : array, shape (7,)
        Goal position.

    goal_yd : array, shape (6,)
        Goal velocity.

    goal_ydd : array, shape (6,)
        Goal acceleration.

    start_y : array, shape (7,)
        Start position.

    start_yd : array, shape (6,)
        Start velocity.

    start_ydd : array, shape (6,)
        Start acceleration.

    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    alpha_y : array, shape (6,)
        Constant in transformation system.

    beta_y : array, shape (6,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    coupling_term_precomputed : tuple
        A precomputed coupling term, i.e., additional velocity and
        acceleration.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    Raises
    ------
    ValueError
        If goal time is before start time.
    """
    if start_t >= goal_t:
        raise ValueError("Goal must be chronologically after start!")

    if t <= start_t:
        return np.copy(start_y), np.copy(start_yd), np.copy(start_ydd)

    execution_time = goal_t - start_t

    current_ydd = np.empty_like(current_yd)

    current_t = last_t
    while current_t < t:
        dt = int_dt
        if t - current_t < int_dt:
            dt = t - current_t
        current_t += dt

        if coupling_term is not None:
            cd, cdd = coupling_term.coupling(current_y, current_yd)
        else:
            cd, cdd = np.zeros(3), np.zeros(3)
        if coupling_term_precomputed is not None:
            cd += coupling_term_precomputed[0]
            cdd += coupling_term_precomputed[1]

        z = forcing_term.phase(current_t)
        f = forcing_term.forcing_term(z).squeeze()

        if smooth_scaling:
            goal_y_minus_start_y = pr.compact_axis_angle_from_quaternion(
                pr.concatenate_quaternions(goal_y, pr.q_conj(start_y)))
            smoothing = beta_y * z * goal_y_minus_start_y
        else:
            smoothing = 0.0

        current_ydd[:] = (
            alpha_y * (
                beta_y * pr.compact_axis_angle_from_quaternion(
                    pr.concatenate_quaternions(goal_y, pr.q_conj(current_y))
                )
                - execution_time * current_yd
                - smoothing
            )
            + f
            + cdd
        ) / execution_time ** 2
        current_yd += dt * current_ydd + cd / execution_time
        current_y[:] = pr.concatenate_quaternions(
            pr.quaternion_from_compact_axis_angle(dt * current_yd), current_y)


CARTESIAN_DMP_STEP_FUNCTIONS = {
    "python": dmp_step_quaternion_python
}


try:
    from ..dmp_fast import dmp_step_quaternion
    CARTESIAN_DMP_STEP_FUNCTIONS["cython"] = dmp_step_quaternion
    DEFAULT_CARTESIAN_DMP_STEP_FUNCTION = "cython"
except ImportError:
    DEFAULT_CARTESIAN_DMP_STEP_FUNCTION = "python"


class CartesianDMP(DMPBase):
    r"""Cartesian dynamical movement primitive.

    The Cartesian DMP handles orientation and position separately. The
    orientation is represented by a quaternion.

    While the dimension of the state space is 7, the dimension of the
    velocity, acceleration, and forcing term is 6.

    Equation of transformation system for the orientation (according to [1]_,
    Eq. 16):

    .. math::

        \ddot{y} = (\alpha_y (\beta_y (g - y) - \tau \dot{y}) + f(z) + C_t) / \tau^2

    Note that in this case :math:`y` is a quaternion in this case,
    :math:`g - y` the quaternion difference (expressed as rotation vector),
    :math:`\dot{y}` is the angular velocity, and :math:`\ddot{y}` the
    angular acceleration.

    With smooth scaling (according to [2]_):

    .. math::

        \ddot{y} = (\alpha_y (\beta_y (g - y) - \tau \dot{y}
        - \underline{\beta_y (g - y_0) z}) + f(z) + C_t) / \tau^2

    The position is handled in the same way, just like in the original
    :class:`DMP`.

    Parameters
    ----------
    execution_time : float, optional (default: 1)
        Execution time of the DMP.

    dt : float, optional (default: 0.01)
        Time difference between DMP steps.

    n_weights_per_dim : int, optional (default: 10)
        Number of weights of the function approximator per dimension.

    int_dt : float, optional (default: 0.001)
        Time difference for Euler integration.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    alpha_y : float or array-like, shape (6,), optional (default: 25.0)
        Parameter of the transformation system.

    beta_y : float or array-like, shape (6,), optional (default: 6.25)
        Parameter of the transformation system.

    Attributes
    ----------
    execution_time_ : float
        Execution time of the DMP.

    dt_ : float
        Time difference between DMP steps. This value can be changed to adapt
        the frequency.

    References
    ----------
    .. [1] Ude, A., Nemec, B., Petric, T., Morimoto, J. (2014).
       Orientation in Cartesian space dynamic movement primitives.
       In IEEE International Conference on Robotics and Automation (ICRA)
       (pp. 2997-3004). DOI: 10.1109/ICRA.2014.6907291,
       https://acat-project.eu/modules/BibtexModule/uploads/PDF/udenemecpetric2014.pdf

    .. [2] Pastor, P., Hoffmann, H., Asfour, T., Schaal, S. (2009). Learning
       and Generalization of Motor Skills by Learning from Demonstration.
       In 2009 IEEE International Conference on Robotics and Automation,
       (pp. 763-768). DOI: 10.1109/ROBOT.2009.5152385,
       https://h2t.iar.kit.edu/pdf/Pastor2009.pdf
    """
    def __init__(
            self, execution_time=1.0, dt=0.01, n_weights_per_dim=10,
            int_dt=0.001, smooth_scaling=False, alpha_y=25.0, beta_y=6.25):
        super(CartesianDMP, self).__init__(7, 6)
        self._execution_time = execution_time
        self.dt_ = dt
        self.n_weights_per_dim = n_weights_per_dim
        self.int_dt = int_dt
        self.smooth_scaling = smooth_scaling

        self._init_forcing_term()

        self.alpha_y = ensure_1d_array(alpha_y, 6, "alpha_y")
        self.beta_y = ensure_1d_array(beta_y, 6, "beta_y")

    def _init_forcing_term(self):
        alpha_z = canonical_system_alpha(0.01, self.execution_time_, 0.0)
        self.forcing_term_pos = ForcingTerm(
            3, self.n_weights_per_dim, self.execution_time_, 0.0, 0.8,
            alpha_z)
        self.forcing_term_rot = ForcingTerm(
            3, self.n_weights_per_dim, self.execution_time_, 0.0, 0.8,
            alpha_z)

    def get_execution_time_(self):
        return self._execution_time

    def set_execution_time_(self, execution_time):
        self._execution_time = execution_time
        weights_pos = self.forcing_term_pos.weights_
        weights_rot = self.forcing_term_rot.weights_
        self._init_forcing_term()
        self.forcing_term_pos.weights_ = weights_pos
        self.forcing_term_rot.weights_ = weights_rot

    execution_time_ = property(get_execution_time_, set_execution_time_)

    def step(self, last_y, last_yd, coupling_term=None,
             step_function=DMP_STEP_FUNCTIONS[DEFAULT_DMP_STEP_FUNCTION],
             quaternion_step_function=CARTESIAN_DMP_STEP_FUNCTIONS[
                 DEFAULT_CARTESIAN_DMP_STEP_FUNCTION]):
        """DMP step.

        Parameters
        ----------
        last_y : array, shape (7,)
            Last state.

        last_yd : array, shape (6,)
            Last time derivative of state (velocity).

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        step_function : callable, optional (default: RK4)
            DMP integration function.

        quaternion_step_function : callable, optional (default: cython code if available)
            DMP integration function.

        Returns
        -------
        y : array, shape (14,)
            Next state.

        yd : array, shape (12,)
            Next time derivative of state (velocity).
        """
        assert len(last_y) == 7
        assert len(last_yd) == 6

        self.last_t = self.t
        self.t += self.dt_

        # TODO tracking error

        self.current_y[:], self.current_yd[:] = last_y, last_yd
        step_function(
            self.last_t, self.t,
            self.current_y[:3], self.current_yd[:3],
            self.goal_y[:3], self.goal_yd[:3], self.goal_ydd[:3],
            self.start_y[:3], self.start_yd[:3], self.start_ydd[:3],
            self.execution_time_, 0.0,
            self.alpha_y[:3], self.beta_y[:3],
            self.forcing_term_pos,
            coupling_term=coupling_term,
            int_dt=self.int_dt,
            smooth_scaling=self.smooth_scaling)
        quaternion_step_function(
            self.last_t, self.t,
            self.current_y[3:], self.current_yd[3:],
            self.goal_y[3:], self.goal_yd[3:], self.goal_ydd[3:],
            self.start_y[3:], self.start_yd[3:], self.start_ydd[3:],
            self.execution_time_, 0.0,
            self.alpha_y[3:], self.beta_y[3:],
            self.forcing_term_rot,
            coupling_term=coupling_term,
            int_dt=self.int_dt,
            smooth_scaling=self.smooth_scaling)
        return np.copy(self.current_y), np.copy(self.current_yd)

    def open_loop(self, run_t=None, coupling_term=None,
                  step_function=DEFAULT_DMP_STEP_FUNCTION,
                  quaternion_step_function=DEFAULT_CARTESIAN_DMP_STEP_FUNCTION):
        """Run DMP open loop.

        Parameters
        ----------
        run_t : float, optional (default: execution_time)
            Run time of DMP. Can be shorter or longer than execution_time.

        coupling_term : object, optional (default: None)
            Coupling term that will be added to velocity.

        step_function : str, optional (default: 'rk4')
            DMP integration function. Possible options: 'rk4', 'euler',
            'euler-cython', 'rk4-cython'.

        quaternion_step_function : str, optional (default: 'cython' if available)
            DMP integration function. Possible options: 'python', 'cython'.

        Returns
        -------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, 7)
            State at each step.
        """
        try:
            step_function = DMP_STEP_FUNCTIONS[step_function]
        except KeyError:
            raise ValueError(
                f"Step function must be in "
                f"{DMP_STEP_FUNCTIONS.keys()}.")
        T, Yp = dmp_open_loop(
            self.execution_time_, 0.0, self.dt_,
            self.start_y[:3], self.goal_y[:3],
            self.alpha_y[:3], self.beta_y[:3],
            self.forcing_term_pos,
            coupling_term,
            run_t, self.int_dt,
            step_function=step_function,
            smooth_scaling=self.smooth_scaling)
        try:
            quaternion_step_function = CARTESIAN_DMP_STEP_FUNCTIONS[
                quaternion_step_function]
        except KeyError:
            raise ValueError(
                f"Step function must be in "
                f"{CARTESIAN_DMP_STEP_FUNCTIONS.keys()}.")
        _, Yr = dmp_open_loop_quaternion(
            self.execution_time_, 0.0, self.dt_,
            self.start_y[3:], self.goal_y[3:],
            self.alpha_y[3:], self.beta_y[3:],
            self.forcing_term_rot,
            coupling_term,
            run_t, self.int_dt,
            quaternion_step_function,
            self.smooth_scaling)
        return T, np.hstack((Yp, Yr))

    def imitate(self, T, Y, regularization_coefficient=0.0,
                allow_final_velocity=False):
        """Imitate demonstration.

        Target forces for the forcing term are computed for the positions
        in a similar way as in :func:`DMP.imitate`. For the orientations
        we adapt this to handle quaternions adequately.

        Parameters
        ----------
        T : array, shape (n_steps,)
            Time for each step.

        Y : array, shape (n_steps, 7)
            State at each step.

        regularization_coefficient : float, optional (default: 0)
            Regularization coefficient for regression.

        allow_final_velocity : bool, optional (default: False)
            Allow a final velocity.
        """
        self.forcing_term_pos.weights_[:, :] = dmp_imitate(
            T, Y[:, :3],
            n_weights_per_dim=self.n_weights_per_dim,
            regularization_coefficient=regularization_coefficient,
            alpha_y=self.alpha_y[:3], beta_y=self.beta_y[:3],
            overlap=self.forcing_term_pos.overlap,
            alpha_z=self.forcing_term_pos.alpha_z,
            allow_final_velocity=allow_final_velocity,
            smooth_scaling=self.smooth_scaling)[0]
        self.forcing_term_rot.weights_[:, :] = dmp_quaternion_imitation(
            T, Y[:, 3:],
            n_weights_per_dim=self.n_weights_per_dim,
            regularization_coefficient=regularization_coefficient,
            alpha_y=self.alpha_y[3:], beta_y=self.beta_y[3:],
            overlap=self.forcing_term_rot.overlap,
            alpha_z=self.forcing_term_rot.alpha_z,
            allow_final_velocity=allow_final_velocity,
            smooth_scaling=self.smooth_scaling)[0]

        self.configure(start_y=Y[0], goal_y=Y[-1])

    def get_weights(self):
        """Get weight vector of DMP.

        Returns
        -------
        weights : array, shape (6 * n_weights_per_dim,)
            Current weights of the DMP.
        """
        return np.concatenate((self.forcing_term_pos.weights_.ravel(),
                               self.forcing_term_rot.weights_.ravel()))

    def set_weights(self, weights):
        """Set weight vector of DMP.

        Parameters
        ----------
        weights : array, shape (6 * n_weights_per_dim,)
            New weights of the DMP.
        """
        n_pos_weights = self.forcing_term_pos.weights_.size
        self.forcing_term_pos.weights_[:, :] = weights[:n_pos_weights].reshape(
            -1, self.n_weights_per_dim)
        self.forcing_term_rot.weights_[:, :] = weights[n_pos_weights:].reshape(
            -1, self.n_weights_per_dim)


def dmp_quaternion_imitation(
        T, Y, n_weights_per_dim, regularization_coefficient, alpha_y, beta_y,
        overlap, alpha_z, allow_final_velocity, smooth_scaling=False):
    """Compute weights and metaparameters of quaternion DMP.

    Parameters
    ----------
    T : array, shape (n_steps,)
        Time of each step.

    Y : array, shape (n_steps, 4)
        Orientation at each step.

    n_weights_per_dim : int
        Number of weights per dimension.

    regularization_coefficient : float, optional (default: 0)
        Regularization coefficient for regression.

    alpha_y : array, shape (3,)
        Parameter of the transformation system.

    beta_y : array, shape (3,)
        Parameter of the transformation system.

    overlap : float
        At which value should radial basis functions of the forcing term
        overlap?

    alpha_z : float
        Parameter of the canonical system.

    allow_final_velocity : bool
        Whether a final velocity is allowed. Will be set to 0 otherwise.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    Returns
    -------
    weights : array, shape (3, n_weights_per_dim)
        Weights of the forcing term.

    start_y : array, shape (4,)
        Start orientation.

    start_yd : array, shape (3,)
        Start velocity.

    start_ydd : array, shape (3,)
        Start acceleration.

    goal_y : array, shape (4,)
        Final orientation.

    goal_yd : array, shape (3,)
        Final velocity.

    goal_ydd : array, shape (3,)
        Final acceleration.
    """
    if regularization_coefficient < 0.0:
        raise ValueError("Regularization coefficient must be >= 0!")

    forcing_term = ForcingTerm(
        3, n_weights_per_dim, T[-1], T[0], overlap, alpha_z)
    F, start_y, start_yd, start_ydd, goal_y, goal_yd, goal_ydd = \
        determine_forces_quaternion(
            T, Y, alpha_y, beta_y, alpha_z, allow_final_velocity,
            smooth_scaling)  # n_steps x n_dims

    X = forcing_term.design_matrix(T)  # n_weights_per_dim x n_steps

    return (ridge_regression(X, F, regularization_coefficient),
            start_y, start_yd, start_ydd, goal_y, goal_yd, goal_ydd)


def determine_forces_quaternion(
        T, Y, alpha_y, beta_y, alpha_z, allow_final_velocity,
        smooth_scaling=False):
    """Determine forces that the forcing term should generate.

    Parameters
    ----------
    T : array, shape (n_steps,)
        Time of each step.

    Y : array, shape (n_steps, n_dims)
        Position at each step.

    alpha_y : array, shape (6,)
        Parameter of the transformation system.

    beta_y : array, shape (6,)
        Parameter of the transformation system.

    alpha_z : float
        Parameter of the canonical system.

    allow_final_velocity : bool
        Whether a final velocity is allowed. Will be set to 0 otherwise.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    Returns
    -------
    F : array, shape (n_steps, n_dims)
        Forces.

    start_y : array, shape (4,)
        Start orientation.

    start_yd : array, shape (3,)
        Start velocity.

    start_ydd : array, shape (3,)
        Start acceleration.

    goal_y : array, shape (4,)
        Final orientation.

    goal_yd : array, shape (3,)
        Final velocity.

    goal_ydd : array, shape (3,)
        Final acceleration.
    """
    n_dims = 3

    DT = np.gradient(T)

    Yd = pr.quaternion_gradient(Y) / DT[:, np.newaxis]
    if not allow_final_velocity:
        Yd[-1, :] = 0.0

    Ydd = np.empty_like(Yd)
    for d in range(n_dims):
        Ydd[:, d] = np.gradient(Yd[:, d]) / DT
    Ydd[-1, :] = 0.0

    execution_time = T[-1] - T[0]
    goal_y = Y[-1]
    start_y = Y[0]
    goal_y_minus_start_y = pr.compact_axis_angle_from_quaternion(
        pr.concatenate_quaternions(goal_y, pr.q_conj(start_y)))
    S = phase(T, alpha_z, T[-1], T[0])
    F = np.empty((len(T), n_dims))
    for t in range(len(T)):
        if smooth_scaling:
            smoothing = beta_y * S[t] * goal_y_minus_start_y
        else:
            smoothing = 0.0
        F[t, :] = execution_time ** 2 * Ydd[t] - alpha_y * (
            beta_y * pr.compact_axis_angle_from_quaternion(
                pr.concatenate_quaternions(goal_y, pr.q_conj(Y[t])))
            - execution_time * Yd[t]
            - smoothing
        )
    return F, Y[0], Yd[0], Ydd[0], Y[-1], Yd[-1], Ydd[-1]


def dmp_open_loop_quaternion(
        goal_t, start_t, dt, start_y, goal_y, alpha_y, beta_y, forcing_term,
        coupling_term=None, run_t=None, int_dt=0.001,
        quaternion_step_function=CARTESIAN_DMP_STEP_FUNCTIONS[
            DEFAULT_CARTESIAN_DMP_STEP_FUNCTION],
        smooth_scaling=False):
    """Run Cartesian DMP without external feedback.

    Parameters
    ----------
    goal_t : float
        Time at the end.

    start_t : float
        Time at the start.

    dt : float, optional (default: 0.01)
        Time difference between DMP steps.

    start_y : array, shape (7,)
        Start position.

    goal_y : array, shape (7,)
        Goal position.

    alpha_y : array, shape (6,)
        Constant in transformation system.

    beta_y : array, shape (6,)
        Constant in transformation system.

    forcing_term : ForcingTerm
        Forcing term.

    coupling_term : CouplingTerm, optional (default: None)
        Coupling term. Must have a function coupling(y, yd) that returns
        additional velocity and acceleration.

    run_t : float, optional (default: goal_t)
        Time at which the DMP will be stopped.

    int_dt : float, optional (default: 0.001)
        Time delta used internally for integration.

    quaternion_step_function : callable, optional (default: cython code if available)
        DMP integration function.

    smooth_scaling : bool, optional (default: False)
        Avoids jumps during the beginning of DMP execution when the goal
        is changed and the trajectory is scaled by interpolating between
        the old and new scaling of the trajectory.

    Returns
    -------
    T : array, shape (n_steps,)
        Times.

    Y : array, shape (n_steps, 4)
        Orientations.
    """
    goal_yd = np.zeros(3)
    goal_ydd = np.zeros(3)
    start_yd = np.zeros(3)
    start_ydd = np.zeros(3)

    if run_t is None:
        run_t = goal_t

    current_y = np.copy(start_y)
    current_yd = np.copy(start_yd)

    T = np.arange(start_t, run_t + dt, dt)
    Y = np.empty((len(T), len(current_y)))
    Y[0] = current_y

    for i in range(1, len(T)):
        quaternion_step_function(
            T[i - 1], T[i], current_y, current_yd,
            goal_y=goal_y, goal_yd=goal_yd, goal_ydd=goal_ydd,
            start_y=start_y, start_yd=start_yd, start_ydd=start_ydd,
            goal_t=goal_t, start_t=start_t,
            alpha_y=alpha_y, beta_y=beta_y,
            forcing_term=forcing_term, coupling_term=coupling_term,
            int_dt=int_dt, smooth_scaling=smooth_scaling)
        Y[i] = current_y

    return T, Y



import numpy as np
import pytransform3d.visualizer as pv
from pytransform3d.urdf import UrdfTransformManager
from mocap.dataset_loader import load_kuka_dataset
from movement_primitives.promp import ProMP


class SelectDemo:
    def __init__(self, fig, left, right):
        self.fig = fig
        self.left = left
        self.right = right
        self.show = True

    def __call__(self, vis, key, modifier):  # sometimes we don't receive the correct keys, why?
        self.show = not self.show
        if self.show and modifier:
            for g in self.left.geometries + self.right.geometries:
                vis.remove_geometry(g, False)
        elif not self.show and not modifier:
            for g in self.left.geometries + self.right.geometries:
                vis.add_geometry(g, False)
        return True


n_weights_per_dim = 10

pattern = "data/kuka/20200129_peg_in_hole/csv_processed/01_peg_in_hole_both_arms/*.csv"
#pattern = "data/kuka/20191213_carry_heavy_load/csv_processed/01_heavy_load_no_tilt_0cm_dual_arm/*.csv"
#pattern = "data/kuka/20191023_rotate_panel_varying_size/csv_processed/panel_450mm_counterclockwise/*.csv"

dataset = load_kuka_dataset(pattern, verbose=1)
Ts = [entry[0] for entry in dataset[:5]]
Ps = [entry[1] for entry in dataset[:5]]

mean_n_steps = int(np.mean([len(P) for P in Ps]))
mean_T = np.linspace(0, 1, mean_n_steps)
promp = ProMP(n_dims=14, n_weights_per_dim=n_weights_per_dim)
promp.imitate(Ts, Ps)

fig = pv.figure(with_key_callbacks=True)
fig.plot_transform(s=0.1)
tm = UrdfTransformManager()
with open("kuka_lbr/urdf/kuka_lbr.urdf", "r") as f:
    tm.load_urdf(f.read(), mesh_path="kuka_lbr/urdf/")
fig.plot_graph(tm, "kuka_lbr", show_visuals=True)

# sample trajectories from ProMP
random_state = np.random.RandomState(0)
for idx in range(10):
    print("Reproduce #%d" % (idx + 1))

    P = promp.sample_trajectories(mean_T, 1, random_state)[0]

    left = fig.plot_trajectory(P=P[:, :7], s=0.02)
    right = fig.plot_trajectory(P=P[:, 7:], s=0.02)
    key = ord(str((idx + 1) % 10))
    fig.visualizer.register_key_action_callback(key, SelectDemo(fig, left, right))

fig.view_init(azim=0, elev=25)
fig.show()



import numpy as np
from movement_primitives.dmp import DualCartesianDMP, CouplingTermDualCartesianPose
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
import pytransform3d.visualizer as vis
from pytransform3d.urdf import UrdfTransformManager
from movement_primitives.testing.simulation import SimulationMockup

dt = 0.001
int_dt = 0.001
execution_time = 1.0

desired_distance = np.array([  # right arm to left arm
    [1.0, 0.0, 0.0, 0.0],
    [0.0, 1.0, 0.0, 0.0],
    [0.0, 0.0, 1.0, -1.2],
    [0.0, 0.0, 0.0, 1.0]
])
desired_distance[:3, :3] = pr.matrix_from_compact_axis_angle([np.deg2rad(180), 0, 0])
ct = CouplingTermDualCartesianPose(desired_distance=desired_distance, couple_position=True, couple_orientation=True, lf=(1.0, 0.0), k=1, c1=0.1, c2=10000)  # c2=10000 in simulation

rh5 = SimulationMockup(dt=dt)

Y = np.zeros((1001, 14))
T = np.linspace(0, 1, len(Y))
sigmoid = 0.5 * (np.tanh(1.5 * np.pi * (T - 0.5)) + 1.0)
radius = 0.5

circle1 = radius * np.cos(np.deg2rad(90) + np.deg2rad(90) * sigmoid)
circle2 = radius * np.sin(np.deg2rad(90) + np.deg2rad(90) * sigmoid)
Y[:, 0] = circle1
Y[:, 1] = 0.55
Y[:, 2] = circle2
R_three_fingers_front = pr.matrix_from_axis_angle([0, 0, 1, 0.5 * np.pi])
R_to_center_start = pr.matrix_from_axis_angle([1, 0, 0, np.deg2rad(0)])
# introduces coupling error (default goal: -90; error at: -110)
R_to_center_end = pr.matrix_from_axis_angle([1, 0, 0, np.deg2rad(-110)])
q_start = pr.quaternion_from_matrix(R_three_fingers_front.dot(R_to_center_start))
q_end = -pr.quaternion_from_matrix(R_three_fingers_front.dot(R_to_center_end))
for i, t in enumerate(sigmoid):
    Y[i, 3:7] = pr.quaternion_slerp(q_start, q_end, t)

circle1 = radius * np.cos(np.deg2rad(270) + np.deg2rad(90) * sigmoid)
circle2 = radius * np.sin(np.deg2rad(270) + np.deg2rad(90) * sigmoid)
Y[:, 7] = circle1
Y[:, 8] = 0.55
Y[:, 9] = circle2
R_three_fingers_front = pr.matrix_from_axis_angle([0, 0, 1, 0.5 * np.pi])
R_to_center_start = pr.matrix_from_axis_angle([1, 0, 0, np.deg2rad(-180)])
R_to_center_end = pr.matrix_from_axis_angle([1, 0, 0, np.deg2rad(-270)])
q_start = pr.quaternion_from_matrix(R_three_fingers_front.dot(R_to_center_start))
q_end = pr.quaternion_from_matrix(R_three_fingers_front.dot(R_to_center_end))
for i, t in enumerate(sigmoid):
    Y[i, 10:] = pr.quaternion_slerp(q_start, q_end, t)


dmp = DualCartesianDMP(
    execution_time=execution_time, dt=dt,
    n_weights_per_dim=10, int_dt=int_dt, p_gain=0.0)
dmp.imitate(T, Y)
dmp.configure(start_y=Y[0], goal_y=Y[-1])

recorded_trajectories = []
lf = (1.0, 0.0)
controller_args = dict(k=1, c1=0.1, c2=10000, verbose=0)
couple_both = CouplingTermDualCartesianPose(
    desired_distance=desired_distance, couple_position=True, couple_orientation=True,
    lf=lf, **controller_args)
couple_orientation = CouplingTermDualCartesianPose(
    desired_distance=desired_distance, couple_position=False, couple_orientation=True,
    lf=lf, **controller_args)
couple_position = CouplingTermDualCartesianPose(
    desired_distance=desired_distance, couple_position=True, couple_orientation=False,
    lf=lf, **controller_args)
for coupling_term in [couple_both, couple_orientation, couple_position, None]:
    dmp.reset()

    rh5.goto_ee_state(Y[0])
    desired_positions, positions, desired_velocities, velocities = \
        rh5.step_through_cartesian(dmp, Y[0], np.zeros(12), execution_time, coupling_term=coupling_term)
    P = np.asarray(positions)

    recorded_trajectories.append(P)

tm = UrdfTransformManager(check=False)
with open("abstract-urdf-gripper/urdf/GripperLeft.urdf", "r") as f:
    tm.load_urdf(f.read(), mesh_path="abstract-urdf-gripper/urdf/")
with open("abstract-urdf-gripper/urdf/GripperRight.urdf", "r") as f:
    tm.load_urdf(f.read(), mesh_path="abstract-urdf-gripper/urdf/")
tm.add_transform("ALWristPitch_Link", "base", np.eye(4))
tm.add_transform("ARWristPitch_Link", "base", np.eye(4))
# assert tm.check_consistency()

fig = vis.figure()
fig.plot_basis(R=np.eye(3), s=0.1)
for P, c in zip(recorded_trajectories, [[0, 1, 0], [1, 0, 0], [1, 0.5, 0], [0, 0, 0]]):
    fig.plot_trajectory(P=P[:, :7], s=0.05, c=c)
    fig.plot_trajectory(P=P[:, 7:], s=0.05, c=c)
graph = fig.plot_graph(tm, "base", show_visuals=True, show_frames=True,
                       whitelist=["ALWristPitch_Link", "ARWristPitch_Link"], s=0.03)


def animation_callback(t, recorded_trajectories, tm, graph):
    k = t // (len(recorded_trajectories[0]) // 10)
    P = recorded_trajectories[k]
    t = t % (len(recorded_trajectories[0]) // 10)
    gripper_left2base = pt.transform_from_pq(P[10 * t, :7])
    gripper_right2base = pt.transform_from_pq(P[10 * t, 7:])
    tm.add_transform("ALWristPitch_Link", "base", gripper_left2base)
    tm.add_transform("ARWristPitch_Link", "base", gripper_right2base)
    graph.set_data()
    return graph


fig.view_init(elev=0, azim=90)
fig.animate(
    animation_callback, len(recorded_trajectories) * len(recorded_trajectories[0]) // 10,
    loop=True, fargs=(recorded_trajectories, tm, graph))
fig.show()



import numpy as np
from movement_primitives.dmp import DMP
from functools import partial
import timeit


n_dims = 10
start_y = np.zeros(n_dims)
goal_y = np.ones(n_dims)
dt = 0.001
int_dt = 0.0001

dmp = DMP(n_dims=n_dims, execution_time=1.0, dt=dt, n_weights_per_dim=6, int_dt=int_dt)
dmp.configure(start_y=start_y, goal_y=goal_y)
dmp.set_weights(1000 * np.random.randn(*dmp.get_weights().shape))

times = timeit.repeat(partial(dmp.open_loop, step_function="rk4"), repeat=10, number=1)
print("RK4")
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))

times = timeit.repeat(partial(dmp.open_loop, step_function="euler"), repeat=10, number=1)
print("Euler (Python)")
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))

times = timeit.repeat(partial(dmp.open_loop, step_function="euler-cython"), repeat=10, number=1)
print("Euler (Cython)")
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))

times = timeit.repeat(partial(dmp.open_loop, step_function="rk4-cython"), repeat=10, number=1)
print("RK4 (Cython)")
print("Mean: %.5f; Std. dev.: %.5f" % (np.mean(times), np.std(times)))



#!/usr/bin/env python
import os
import warnings
from setuptools import setup, find_packages


if __name__ == "__main__":
    with open("README.md", "r") as f:
        long_description = f.read()
    with open(os.path.join("movement_primitives", "_version.py")) as f:
        version = f.read().strip().split('"')[1]
    setup_config = dict(
        name="movement_primitives",
        version=version,
        author="Alexander Fabisch",
        author_email="alexander.fabisch@dfki.de",
        url="https://github.com/dfki-ric/movement_primitives",
        description="Movement primitives",
        long_description=long_description,
        long_description_content_type="text/markdown",
        classifiers=["Programming Language :: Python :: 3"],
        license="BSD-3-clause",
        packages=find_packages(),
        install_requires=[],
        extras_require={
            "all": ["pytransform3d", "cython", "numpy", "scipy", "matplotlib",
                    "open3d", "tqdm", "gmr", "PyYAML", "numba", "pybullet"],
            "doc": ["sphinx", "pydata-sphinx-theme", "numpydoc", "myst-parser"],
            "test": ["pytest", "pytest-cov"]
        }
    )
    try:
        from Cython.Build import cythonize
        import numpy
        cython_config = dict(
            ext_modules=cythonize("movement_primitives/dmp_fast.pyx"),
            zip_safe=False,
            compiler_directives={"language_level": "3"},
            include_dirs=[numpy.get_include()],
            extra_compile_args=[
                "-O3",
                "-Wno-cpp", "-Wno-unused-function"
            ]
        )
        setup_config.update(cython_config)
    except ImportError:
        warnings.warn("Cython or NumPy is not available. "
                      "Install it if you want fast DMPs.")

    setup(**setup_config)
